from time import time
from typing import Any
from multiprocessing import context, connection, get_context
import sys

from .pickle import dumps, loads


class ConnectionWrapper(object):  # pragma: no cover
    """
    This simple wrapper provides timeout function for sending
    bytes on ``Connection``.
    """

    def __init__(self, conn):
        self.conn = conn

    def send_bytes(self, bytes_):
        """
        Send bytes over the connection pipe.
        """
        self.conn.send_bytes(bytes_)

    def recv_bytes(self, timeout=None):
        """
        Receive bytes from the connection pipe.

        Raises:
            TimeoutError if timeout.
        """
        if self.conn.poll(timeout=timeout):
            return self.conn.recv_bytes()
        else:
            raise TimeoutError("Timeout")

    def __getattr__(self, name):
        if 'conn' in self.__dict__:
            return getattr(self.conn, name)
        raise AttributeError("'{}' object has no attribute '{}'".format(
            type(self).__name__, 'conn'))


class SimpleQueue(object):  # pragma: no cover
    """
    A simple single direction queue for inter-process communications.
    There could be multiple receivers and multiple senders on each side.
    """
    def __init__(self, *, ctx=None, copy_tensor=False):
        """
        Args:
            ctx: Multiprocessing context, you can get this using ``get_context``
            copy_tensor: Set the queue to send a fully serialized tensor
                if ``True``, and only a stub of reference if ``False``.

        See Also:
            :func:`.dump_tensor`
        """
        if ctx is None:
            # get default context
            ctx = get_context()
        self._reader, self._writer = connection.Pipe(duplex=False)
        self._reader = ConnectionWrapper(self._reader)
        self._writer = ConnectionWrapper(self._writer)
        # _rlock will be used by _help_stuff_finish() of multiprocessing.Pool
        self._rlock = ctx.Lock()
        self._copy_tensor = copy_tensor
        if sys.platform == 'win32':
            self._wlock = None
        else:
            self._wlock = ctx.Lock()

    def empty(self):
        """
        Returns:
            Whether the queue is empty or not.
        """
        return not self._reader.poll()

    def close(self):
        self._reader.close()
        self._writer.close()

    def __getstate__(self):
        context.assert_spawning(self)
        return (self._reader, self._writer, self._rlock, self._wlock,
                self._copy_tensor)

    def __setstate__(self, state):
        (self._reader, self._writer, self._rlock, self._wlock,
         self._copy_tensor) = state

    def get(self, timeout=None):
        """
        Get an object from the queue. This api is required by
        ``multiprocessing.pool`` to perform inter-process
        communication.

        Note:
            This api is used by sub-processes in pool to get tasks
            and work.

        Returns:
            Any object.
        """
        with self._rlock:
            res = self._reader.recv_bytes(timeout)
        # deserialize the data after having released the lock
        return loads(res)

    def put(self, obj: Any):
        """
        Put an object into the queue. This api is required by
        ``multiprocessing.pool`` to perform inter-process
        communication.

        Note:
            This api is used by sub-processes in pool to put results
            and respond.

        Args:
            obj: Any object.
        """
        # serialize the data before acquiring the lock
        obj = dumps(obj, copy_tensor=self._copy_tensor)
        if self._wlock is None:
            # writes to a message oriented win32 pipe are atomic
            self._writer.send_bytes(obj)
        else:
            with self._wlock:
                self._writer.send_bytes(obj)

    def quick_get(self, timeout=None):
        """
        Get an object from the queue.

        Note:
            this api is used by the result manager (``Pool._result_handler``)
            thread to get results from the queue, since it is single threaded,
            there is no need to use locks, and therefore quicker.

        Returns:
            Any object.
        """
        res = self._reader.recv_bytes(timeout)
        return loads(res)

    def quick_put(self, obj: Any):
        """
        Put an object into the queue.

        Note: this api is used by the pool manager (``Pool._task_handler``)
            thread to put tasks into the queue, since it is single threaded,
            there is no need to use locks, and therefore quicker.

        Args:
            obj: Any object.
        """
        obj = dumps(obj, copy_tensor=self._copy_tensor)
        self._writer.send_bytes(obj)

    def __del__(self):
        self.close()


class SimpleP2PQueue(object):  # pragma: no cover
    """
    A simple single direction queue for inter-process P2P communications.
    Each end only have one process.
    """
    def __init__(self, *, copy_tensor=False):
        """
        Args:
            copy_tensor: Set the queue to send a fully serialized tensor
                if ``True``, and only a stub of reference if ``False``.

        See Also:
            :func:`.dump_tensor`
        """
        self._reader, self._writer = connection.Pipe(duplex=False)
        self._reader = ConnectionWrapper(self._reader)
        self._writer = ConnectionWrapper(self._writer)
        self._copy_tensor = copy_tensor

    def empty(self):
        """
        Returns:
            Used by receiver, Whether the queue is empty or not.
        """
        return not self._reader.poll()

    def close(self):
        self._reader.close()
        self._writer.close()

    def __getstate__(self):
        context.assert_spawning(self)
        return self._reader, self._writer, self._copy_tensor

    def __setstate__(self, state):
        (self._reader, self._writer, self._copy_tensor) = state

    def get(self, timeout=None):
        """
        Get an object from the queue. This api is required by
        ``multiprocessing.pool`` to perform inter-process
        communication.

        Note:
            This api is used by sub-processes in pool to get tasks
            and work.

        Returns:
            Any object.
        """
        res = self._reader.recv_bytes(timeout)
        # deserialize the data after having released the lock
        return loads(res)

    def put(self, obj: Any):
        """
        Put an object into the queue. This api is required by
        ``multiprocessing.pool`` to perform inter-process
        communication.

        Note:
            This api is used by sub-processes in pool to put results
            and respond.

        Args:
            obj: Any object.
        """
        # serialize the data before acquiring the lock
        obj = dumps(obj, copy_tensor=self._copy_tensor)
        self._writer.send_bytes(obj)

    def __del__(self):
        self.close()


class MultiP2PQueue(object):
    """
    P2P queue which connects pool result manager and worker processes directly,
    with no lock.
    """
    def __init__(self, queue_num, *, copy_tensor=False):
        self.counter = 0
        self.queues = [SimpleP2PQueue(copy_tensor=copy_tensor)
                       for _ in range(queue_num)]

    def put(self, obj: Any):
        # randomly choose a worker's queue
        queue = self.queues[self.counter]
        self.counter = (self.counter + 1) % len(self.queues)
        queue.put(obj)

    def get(self, timeout=None):
        begin = time()
        while True:
            if timeout is not None and time() - begin > timeout:
                raise TimeoutError("Timeout")
            for queue in self.queues:
                try:
                    obj = queue.get(timeout=1e-3)
                    return obj
                except TimeoutError:
                    continue

    def get_sub_queue(self, index):
        return self.queues[index]

    def __del__(self):
        for q in self.queues:
            q.close()


__all__ = ["SimpleQueue"]



from PIL import Image
from typing import List
from machin.parallel import get_context
import os
import numpy as np
import moviepy.editor as mpy
import matplotlib.pyplot as plt


def show_image(image: np.array,
               show_normalized: bool = True,
               pause_time: float = 0.01,
               title: str = ""):
    """
    Use matplotlib to show a single image. You may repeatedly call this method
    with the same ``title`` argument to show a video or a dynamically changing
    image.

    Args:
        image: A numpy array of shape (H, W, C) or (H, W), and with ``dtype``
            = any float or any int.
            When a frame is float type, its value range should be [0, 1].
            When a frame is integer type, its value range should be [0, 255].
        show_normalized: Show normalized image alongside the original one.
        pause_time: Pause time between displaying current image and the next
            one.
        title: Title of the display window.
    """
    if np.issubdtype(image.dtype, np.integer):
        image = image.astype(np.floating) / 255

    fig = plt.figure(title, clear=True)
    fig.canvas.set_window_title(title)
    if show_normalized:
        ax = fig.add_subplot("121")
        ax.set_facecolor((0.0, 0.0, 0.0))
        ax.imshow(image, vmin=np.min(image), vmax=np.max(image))

        ax2 = fig.add_subplot("122")
        ax2.set_facecolor((0.0, 0.0, 0.0))
        pix_range = (np.max(image) - np.min(image)) + 1e-6
        ax2.imshow((image - np.min(image)) / pix_range, vmin=0, vmax=1)
        plt.pause(pause_time)

    else:
        ax = fig.add_subplot("111")
        ax.set_facecolor((0.0, 0.0, 0.0))
        ax.imshow(image, vmin=np.min(image), vmax=np.max(image))
        plt.pause(pause_time)


def create_video(frames: List[np.array],
                 path: str,
                 filename: str,
                 extension: str = ".gif",
                 fps: int = 15):
    """
    Args:
        frames: A list of numpy arrays of shape (H, W, C) or (H, W), and with
            ``dtype`` = any float or any int.
            When a frame is float type, its value range should be [0, 1].
            When a frame is integer type, its value range should be [0, 255].
        path: Directory to save the video.
        filename: File name.
        extension: File extension.
        fps: frames per second.
    """
    if frames:
        for f in range(len(frames)):
            if np.issubdtype(frames[f].dtype, np.integer):
                frames[f] = frames[f].astype(np.uint8)
            elif np.issubdtype(frames[f].dtype, np.floating):
                frames[f] = (frames[f] * 255).astype(np.uint8)
            if frames[f].ndim == 2:
                # consider as a grey scale image
                frames[f] = np.repeat(frames[f][:, :, np.newaxis], 3, axis=2)

        clip = mpy.ImageSequenceClip(frames, fps=fps)
        if extension.lower() == ".gif":
            clip.write_gif(os.path.join(path, filename + extension),
                           fps=fps, verbose=False, logger=None)
        else:
            clip.write_videofile(os.path.join(path, filename + extension),
                                 fps=fps, verbose=False, logger=None)


def create_video_subproc(frames: List[np.array],
                         path: str,
                         filename: str,
                         extension: str = ".gif",
                         fps: int = 15,
                         daemon: bool = True):
    """
    Create video with a subprocess, since it takes a lot of time for ``moviepy``
    to encode the video file.

    See Also:
         :func:`.create_video`

    Note:
        if ``daemon`` is true, then this function cannot be used in a
        daemonic subprocess.

    Args:
        frames: A list of numpy arrays of shape (H, W, C) or (H, W), and with
            ``dtype`` = any float or any int.
            When a frame is float type, its value range should be [0, 1].
            When a frame is integer type, its value range should be [0, 255].
        path: Directory to save the video.
        filename: File name.
        extension: File extension.
        fps: frames per second.
        daemon: Whether launching the saving process as a daemonic process.

    Returns:
        A wait function, once called, block until creation has finished.
    """
    def wait():
        pass

    if frames:
        p = get_context("spawn").Process(target=create_video,
                                         args=(frames,
                                               path,
                                               filename,
                                               extension,
                                               fps))
        p.daemon = daemon
        p.start()

        def wait():
            p.join()

    return wait


def create_image(image: np.array,
                 path: str,
                 filename: str,
                 extension: str = ".png"):
    """
    Args:
        image: A numpy array of shape (H, W, C) or (H, W), and with
            ``dtype`` = any float or any int.
            When a frame is float type, its value range should be [0, 1].
            When a frame is integer type, its value range should be [0, 255].
        path: Directory to save the image.
        filename: File name.
        extension: File extension.
    """
    if np.issubdtype(image.dtype, np.integer):
        image = image.astype(np.uint8)
    elif np.issubdtype(image.dtype, np.floating):
        image = (image * 255).astype(np.uint8)
    if image.ndim == 2:
        # consider as a grey scale image
        image = np.repeat(image[:, :, np.newaxis], 3, axis=2)
    image = Image.fromarray(image)
    image.save(os.path.join(path, filename + extension))


def create_image_subproc(image: np.array,
                         path: str,
                         filename: str,
                         extension: str = ".png",
                         daemon: bool = True):
    """
    Create image with a subprocess.

    See Also:
         :func:`.create_image`

    Note:
        if ``daemon`` is true, then this function cannot be used in a
        daemonic subprocess.

    Args:
        image: A numpy array of shape (H, W, C) or (H, W), and with
            ``dtype`` = any float or any int.
            When a frame is float type, its value range should be [0, 1].
            When a frame is integer type, its value range should be [0, 255].
        path: Directory to save the image.
        filename: File name.
        extension: File extension.
        daemon: Whether launching the saving process as a daemonic process.

    Returns:
        A wait function, once called, block until creation has finished.
    """
    p = get_context("spawn").Process(target=create_image,
                                     args=(image,
                                           path,
                                           filename,
                                           extension))
    p.daemon = daemon
    p.start()

    def wait():
        p.join()

    return wait



# pylint: disable=wildcard-import, unused-wildcard-import
from .ddpg import *


class TD3(DDPG):
    """
    TD3 framework. Which adds a additional pair of critic and target critic
    network to DDPG.
    """

    _is_top = ["actor", "critic", "critic2",
               "actor_target", "critic_target", "critic2_target"]
    _is_restorable = ["actor_target", "critic_target", "critic2_target"]

    def __init__(self,
                 actor: Union[NeuralNetworkModule, nn.Module],
                 actor_target: Union[NeuralNetworkModule, nn.Module],
                 critic: Union[NeuralNetworkModule, nn.Module],
                 critic_target: Union[NeuralNetworkModule, nn.Module],
                 critic2: Union[NeuralNetworkModule, nn.Module],
                 critic2_target: Union[NeuralNetworkModule, nn.Module],
                 optimizer: Callable,
                 criterion: Callable,
                 *_,
                 lr_scheduler: Callable = None,
                 lr_scheduler_args: Tuple[Tuple, Tuple, Tuple] = None,
                 lr_scheduler_kwargs: Tuple[Dict, Dict, Dict] = None,
                 batch_size: int = 100,
                 update_rate: float = 0.005,
                 actor_learning_rate: float = 0.0005,
                 critic_learning_rate: float = 0.001,
                 discount: float = 0.99,
                 gradient_max: float = np.inf,
                 replay_size: int = 500000,
                 replay_device: Union[str, t.device] = "cpu",
                 replay_buffer: Buffer = None,
                 visualize: bool = False,
                 visualize_dir: str = "",
                 **__):
        """
        See Also:
            :class:`.DDPG`

        Args:
            actor: Actor network module.
            actor_target: Target actor network module.
            critic: Critic network module.
            critic_target: Target critic network module.
            critic2: The second critic network module.
            critic2_target: The second target critic network module.
            optimizer: Optimizer used to optimize ``actor``, ``critic``,
            criterion: Criterion used to evaluate the value loss.
            lr_scheduler: Learning rate scheduler of ``optimizer``.
            lr_scheduler_args: Arguments of the learning rate scheduler.
            lr_scheduler_kwargs: Keyword arguments of the learning
                rate scheduler.
            batch_size: Batch size used during training.
            update_rate: :math:`\\tau` used to update target networks.
                Target parameters are updated as:

                :math:`\\theta_t = \\theta * \\tau + \\theta_t * (1 - \\tau)`
            actor_learning_rate: Learning rate of the actor optimizer,
                not compatible with ``lr_scheduler``.
            critic_learning_rate: Learning rate of the critic optimizer,
                not compatible with ``lr_scheduler``.
            discount: :math:`\\gamma` used in the bellman function.
            replay_size: Replay buffer size. Not compatible with
                ``replay_buffer``.
            replay_device: Device where the replay buffer locates on, Not
                compatible with ``replay_buffer``.
            replay_buffer: Custom replay buffer.
            visualize: Whether visualize the network flow in the first pass.
            visualize_dir: Visualized graph save directory.
        """
        if lr_scheduler_args is None:
            lr_scheduler_args = ((), (), ())
        if lr_scheduler_kwargs is None:
            lr_scheduler_kwargs = ({}, {}, {})

        super(TD3, self).__init__(
            actor, actor_target, critic, critic_target, optimizer, criterion,
            lr_scheduler=lr_scheduler,
            lr_scheduler_args=(lr_scheduler_args[:2]
                               if lr_scheduler_args is not None
                               else None),
            lr_scheduler_kwargs=(lr_scheduler_kwargs[:2]
                                 if lr_scheduler_kwargs is not None
                                 else None),
            batch_size=batch_size,
            update_rate=update_rate,
            actor_learning_rate=actor_learning_rate,
            critic_learning_rate=critic_learning_rate,
            discount=discount,
            gradient_max=gradient_max,
            replay_size=replay_size,
            replay_device=replay_device,
            replay_buffer=replay_buffer,
            visualize=visualize,
            visualize_dir=visualize_dir
        )
        self.critic2 = critic2
        self.critic2_target = critic2_target
        self.critic2_optim = optimizer(self.critic2.parameters(),
                                       lr=critic_learning_rate)

        # Make sure target and online networks have the same weight
        with t.no_grad():
            hard_update(self.critic2, self.critic2_target)

        if lr_scheduler is not None:
            self.critic2_lr_sch = lr_scheduler(
                self.critic2_optim,
                *lr_scheduler_args[2],
                **lr_scheduler_kwargs[2]
            )

    def _criticize2(self,
                    state: Dict[str, Any],
                    action: Dict[str, Any],
                    use_target=False,
                    **__):
        """
        Use the second critic network to evaluate current value.

        Args:
            state: Current state.
            action: Current action.
            use_target: Whether to use the target network.

        Returns:
            Q Value of shape ``[batch_size, 1]``.
        """
        if use_target:
            return safe_call(self.critic2_target, state, action)[0]
        else:
            return safe_call(self.critic2, state, action)[0]

    def update(self,
               update_value=True,
               update_policy=True,
               update_target=True,
               concatenate_samples=True,
               **__):
        # DOC INHERITED
        batch_size, (state, action, reward, next_state, terminal, others) = \
            self.replay_buffer.sample_batch(self.batch_size,
                                            concatenate_samples,
                                            sample_method="random_unique",
                                            sample_attrs=[
                                                "state", "action",
                                                "reward", "next_state",
                                                "terminal", "*"
                                            ])

        # Update critic network first.
        # Generate value reference :math: `y_i` using target actor and
        # target critic.
        with t.no_grad():
            next_action = self.action_transform_function(
                self.policy_noise_function(self._act(next_state, True)),
                next_state, others
            )
            next_value = self._criticize(next_state, next_action, True)
            next_value2 = self._criticize2(next_state, next_action, True)
            next_value = t.min(next_value, next_value2)
            next_value = next_value.view(batch_size, -1)
            y_i = self.reward_function(
                reward, self.discount, next_value, terminal, others
            )

        cur_value = self._criticize(state, action)
        cur_value2 = self._criticize2(state, action)
        value_loss = self.criterion(cur_value, y_i.to(cur_value.device))
        value_loss2 = self.criterion(cur_value2, y_i.to(cur_value.device))

        if self.visualize:
            self.visualize_model(value_loss, "critic", self.visualize_dir)

        if update_value:
            self.critic.zero_grad()
            value_loss.backward()
            nn.utils.clip_grad_norm_(
                self.critic.parameters(), self.grad_max
            )
            self.critic_optim.step()
            self.critic2.zero_grad()
            value_loss2.backward()
            nn.utils.clip_grad_norm_(
                self.critic2.parameters(), self.grad_max
            )
            self.critic2_optim.step()

        # Update actor network
        cur_action = self.action_transform_function(
            self._act(state), state, others
        )
        act_value = self._criticize(state, cur_action)

        # "-" is applied because we want to maximize J_b(u),
        # but optimizer workers by minimizing the target
        act_policy_loss = -act_value.mean()

        if self.visualize:
            self.visualize_model(act_policy_loss, "actor", self.visualize_dir)

        if update_policy:
            self.actor.zero_grad()
            act_policy_loss.backward()
            nn.utils.clip_grad_norm_(
                self.actor.parameters(), self.grad_max
            )
            self.actor_optim.step()

        # Update target networks
        if update_target:
            soft_update(self.actor_target, self.actor, self.update_rate)
            soft_update(self.critic_target, self.critic, self.update_rate)
            soft_update(self.critic2_target, self.critic2, self.update_rate)

        # use .item() to prevent memory leakage
        return (-act_policy_loss.item(),
                (value_loss.item() + value_loss2.item()) / 2)

    @staticmethod
    def policy_noise_function(actions, *_):
        # Function used to add noise to actions, mentioned in TD3
        # training tricks
        return actions

    def update_lr_scheduler(self):
        """
        Update learning rate schedulers.
        """
        if hasattr(self, "critic2_lr_sch"):
            self.critic2_lr_sch.step()
        super(TD3, self).update_lr_scheduler()

    def load(self, model_dir: str, network_map: Dict[str, str] = None,
             version: int = -1):
        # DOC INHERITED
        TorchFramework.load(self, model_dir, network_map, version)
        with t.no_grad():
            hard_update(self.actor, self.actor_target)
            hard_update(self.critic, self.critic_target)
            hard_update(self.critic2, self.critic2_target)



from .buffer import Buffer
from .buffer_d import DistributedBuffer
from .prioritized_buffer import WeightTree, PrioritizedBuffer
from .prioritized_buffer_d import DistributedPrioritizedBuffer

__all__ = [
    "Buffer",
    "DistributedBuffer",
    "PrioritizedBuffer",
    "DistributedPrioritizedBuffer",
    "WeightTree"
]



from machin.frame.buffers.prioritized_buffer import PrioritizedBuffer
# pylint: disable=wildcard-import, unused-wildcard-import
from .dqn import *


class RAINBOW(DQN):
    """
    RAINBOW DQN framework.
    """

    def __init__(self,
                 qnet: Union[NeuralNetworkModule, nn.Module],
                 qnet_target: Union[NeuralNetworkModule, nn.Module],
                 optimizer,
                 value_min,
                 value_max,
                 *_,
                 lr_scheduler: Callable = None,
                 lr_scheduler_args: Tuple[Tuple] = None,
                 lr_scheduler_kwargs: Tuple[Dict] = None,
                 batch_size: int = 100,
                 update_rate: float = 0.001,
                 learning_rate: float = 0.001,
                 discount: float = 0.99,
                 gradient_max: float = np.inf,
                 reward_future_steps: int = 3,
                 replay_size: int = 500000,
                 replay_device: Union[str, t.device] = "cpu",
                 replay_buffer: Buffer = None,
                 visualize: bool = False,
                 visualize_dir: str = "",
                 **__):
        """
        RAINBOW framework is described in
        `this <https://arxiv.org/abs/1710.02298>`__ essay.

        Note:
            In the RAINBOW framework, the output shape of your q network
            must be ``[batch_size, action_num, atom_num]`` when given a
            state of shape ``[batch_size, action_dim]``. And the last
            dimension **must be soft-maxed**. Atom number is the number of
            segments of your q value domain.

        See Also:
            :class:`.DQN`

        Args:
            qnet: Q network module.
            qnet_target: Target Q network module.
            optimizer: Optimizer used to optimize ``actor`` and ``critic``.
            value_min: Minimum of value domain.
            value_max: Maximum of value domain.
            learning_rate: Learning rate of the optimizer, not compatible with
                ``lr_scheduler``.
            lr_scheduler: Learning rate scheduler of ``optimizer``.
            lr_scheduler_args: Arguments of the learning rate scheduler.
            lr_scheduler_kwargs: Keyword arguments of the learning
                rate scheduler.
            batch_size: Batch size used during training.
            update_rate: :math:`\\tau` used to update target networks.
                Target parameters are updated as:

                :math:`\\theta_t = \\theta * \\tau + \\theta_t * (1 - \\tau)`

            discount: :math:`\\gamma` used in the bellman function.
            reward_future_steps: Number of future steps to be considered when
                the framework calculates value from reward.
            replay_size: Replay buffer size. Not compatible with
                ``replay_buffer``.
            replay_device: Device where the replay buffer locates on, Not
                compatible with ``replay_buffer``.
            replay_buffer: Custom replay buffer.
            mode: one of ``"vanilla", "fixed_target", "double"``.
            visualize: Whether visualize the network flow in the first pass.
        """
        super(RAINBOW, self).__init__(
            qnet, qnet_target, optimizer, lambda: None,
            learning_rate=learning_rate,
            lr_scheduler=lr_scheduler,
            lr_scheduler_args=lr_scheduler_args,
            lr_scheduler_kwargs=lr_scheduler_kwargs,
            batch_size=batch_size,
            update_rate=update_rate,
            discount=discount,
            gradient_max=gradient_max,
            replay_size=replay_size,
            replay_device=replay_device,
            replay_buffer=(PrioritizedBuffer(replay_size, replay_device)
                           if replay_buffer is None
                           else replay_buffer),
            visualize=visualize,
            visualize_dir=visualize_dir
        )
        self.v_min = value_min
        self.v_max = value_max
        self.reward_future_steps = reward_future_steps

    def act_discrete(self,
                     state: Dict[str, Any],
                     use_target: bool = False,
                     **__):
        # DOC INHERITED
        # q value distribution of each action
        # shape: [batch_size, action_num, atom_num]
        if use_target:
            q_dist, *others = safe_call(self.qnet_target, state)
        else:
            q_dist, *others = safe_call(self.qnet, state)

        atom_num = q_dist.shape[-1]

        # support vector, shape: [1, atom_num]
        q_dist_support = t.linspace(self.v_min, self.v_max, atom_num) \
            .view(1, -1)

        # q value of each action, shape: [batch_size, action_num]
        q_value = t.sum(q_dist_support * q_dist, dim=-1)

        result = t.argmax(q_value, dim=1).view(-1, 1)
        if len(others) == 0:
            return result
        else:
            return (result, *others)

    def act_discrete_with_noise(self,
                                state: Dict[str, Any],
                                use_target: bool = False,
                                **__):
        # DOC INHERITED
        # q value distribution of each action
        # shape: [batch_size, action_num, atom_num]
        if use_target:
            q_dist, *others = safe_call(self.qnet_target, state)
        else:
            q_dist, *others = safe_call(self.qnet, state)

        atom_num = q_dist.shape[-1]

        # support vector, shape: [1, atom_num]
        q_dist_support = t.linspace(self.v_min, self.v_max, atom_num) \
            .view(1, -1)

        # q value of each action, shape: [batch_size, action_num]
        q_value = t.sum(q_dist_support * q_dist, dim=-1)

        result = t.softmax(q_value, dim=1)
        dist = Categorical(result)
        batch_size = result.shape[0]
        sample = dist.sample([batch_size])
        if len(others) == 0:
            return sample
        else:
            return (sample, *others)

    def store_transition(self, transition: Union[Transition, Dict]):
        """
        Add a transition sample to the replay buffer.

        Not suggested, since you will have to calculate "value"
        by yourself.
        """
        self.replay_buffer.append(transition, required_attrs=(
            "state", "action", "next_state", "reward", "value", "terminal"
        ))

    def store_episode(self, episode: List[Union[Transition, Dict]]):
        """
        Add a full episode of transition samples to the replay buffer.

        "value" is automatically calculated.
        """
        episode[-1]["value"] = episode[-1]["reward"]

        # calculate (truncated) n step value for each transition
        for i in reversed(range(len(episode))):
            value_sum = 0
            # for (virtual) transitions beyond the terminal transition,
            # using "min" to ignore them is equivalent to setting their
            # rewards as zero
            for j in reversed(range(min(
                    self.reward_future_steps, len(episode) - i
            ))):
                value_sum = (value_sum * self.discount +
                             episode[i + j]["reward"])
            episode[i]["value"] = value_sum

        for trans in episode:
            self.replay_buffer.append(trans, required_attrs=(
                "state", "action", "next_state", "reward", "value", "terminal"
            ))

    def update(self,
               update_value=True,
               update_target=True,
               concatenate_samples=True,
               **__):
        # DOC INHERITED
        # pylint: disable=invalid-name
        (batch_size,
         (state, action, value, next_state, terminal, others),
         index, is_weight) = \
            self.replay_buffer.sample_batch(self.batch_size,
                                            concatenate_samples,
                                            sample_attrs=[
                                                "state", "action",
                                                "value", "next_state",
                                                "terminal", "*"
                                            ],
                                            additional_concat_attrs=[
                                                "value"
                                            ])

        # q_dist is the distribution of q values
        q_dist = self._criticize(state)
        atom_num = q_dist.shape[-1]

        action = self.action_get_function(action).to(device=q_dist.device,
                                                     dtype=t.long).flatten()
        # shape: [batch_size, atom_num]
        q_dist = q_dist[range(batch_size), action]

        # support vector, shape: [atom_num]
        q_dist_support = t.linspace(self.v_min, self.v_max, atom_num)

        with t.no_grad():
            target_next_q_dist = self._criticize(next_state, True)
            next_action = (self.act_discrete(next_state).flatten()
                           .to(device=target_next_q_dist.device, dtype=t.long))

            # shape: [batch_size, atom_num]
            target_next_q_dist = target_next_q_dist[range(batch_size),
                                                    next_action]

            # shape: [1, atom_num]
            q_dist_support = q_dist_support.unsqueeze(dim=0)

            # T_z is the bellman update for atom z_j
            # shape: [batch_size, atom_num]
            T_z = self.reward_function(
                value,
                self.discount ** self.reward_future_steps,
                q_dist_support,
                terminal,
                others
            )

            # 1e-6 is used to make sure that l != u when T_z == v_min or v_max
            T_z = T_z.clamp(self.v_min + 1e-6, self.v_max - 1e-6)

            # delta_z is the interval length of each atom
            delta_z = (self.v_max - self.v_min) / (atom_num - 1.0)

            # b is the normalized distance of T_z to v_min,
            # l and u are upper and lower atom indexes
            # b, l, u shape: [batch_size, atom_num]
            b = (T_z - self.v_min) / delta_z
            l, u = b.floor(), b.ceil()

            # idx shape: [batch_size * atom_num]
            # dist shape: [batch_size, atom_num]
            # weight shape: [batch_size * atom_num]
            l_idx, l_dist = l.long().view(-1).cpu(), b - l
            u_idx, u_dist = u.long().view(-1).cpu(), u - b
            l_weight = (u_dist * target_next_q_dist).view(-1).cpu()
            u_weight = (l_dist * target_next_q_dist).view(-1).cpu()

            # offset is used to perform row-wise index add, since we can only
            # perform index add on one dimension, we must flatten the whole
            # distribution and then add.
            offset = (t.arange(0, batch_size * atom_num, atom_num)
                      .view(-1, 1)
                      .expand(batch_size, atom_num)
                      .flatten())

            # distribute T_z probability to its nearest upper
            # and lower atom neighbors, using its distance to them.
            # shape: [batch_size * atom_num] -> [batch_size, atom_num]
            # Note: index_add_ on CUDA uses atomicAdd, will cause
            # rounding errors and be a source of noise.
            target_dist = t.zeros([batch_size * atom_num])
            target_dist.index_add_(dim=0, index=l_idx + offset,
                                   source=l_weight)
            target_dist.index_add_(dim=0, index=u_idx + offset,
                                   source=u_weight)
            target_dist = target_dist.view(batch_size, atom_num)

        # target_dist is equivalent to y_i in original dqn
        # division in KL divergence is ignored because target_dist
        # is a constant? But this modification do prevents the 0/0 situation.

        # 1e-6 is used to improve numerical stability and prevent nan
        value_loss = -(target_dist * (q_dist + 1e-6).log().cpu())

        value_loss = value_loss.sum(dim=1)

        abs_error = (t.abs(value_loss) + 1e-6).flatten().detach().numpy()
        self.replay_buffer.update_priority(abs_error, index)

        value_loss = (value_loss *
                      t.from_numpy(is_weight).view([batch_size, 1])).mean()

        if self.visualize:
            self.visualize_model(value_loss, "qnet", self.visualize_dir)

        if update_value:
            self.qnet.zero_grad()
            value_loss.backward()
            nn.utils.clip_grad_norm_(
                self.qnet.parameters(), self.grad_max
            )
            self.qnet_optim.step()

        # Update target Q network
        if update_target:
            soft_update(self.qnet_target, self.qnet, self.update_rate)

        # use .item() to prevent memory leakage
        return value_loss.item()



from typing import Union
import copy
import json
import argparse

from .helper_classes import Object


class Config(Object):
    """
    Attributes:
        root_directory: (const) Root directory of your trials.
        restart_from_trial: (const) Trial name (usually a date string)
            to restart from.
    """

    def __init__(self,
                 root_directory: str = "./",
                 restart_from_trial: str = None,
                 **configs):
        configs["root_directory"] = root_directory
        configs["restart_from_trial"] = restart_from_trial
        super(Config, self).__init__(configs,
                                     const_attrs={"root_directory",
                                                  "restart_from_trial"})


def load_config_cmd(merge_conf: Config = None) -> Config:
    """
    Get configs from the commandline by using "--conf".

    ``--conf a=b`` will set ``<Returned Config>.a = b``

    Example::

        python3 test.py --conf device=\"cuda:1\"
                        --conf some_dict={\"some_key\":1}

    Example::

        from machin.utils.conf import Config
        from machin.utils.save_env import SaveEnv

        # set some config attributes
        c = Config(
            model_save_int = 100,
            root_dir = "some_directory",
            restart_from_trial = "2020_05_09_15_00_31"
        )

        load_config_cmd(c)

        # restart_from_trial specifies the trial name in your root
        # directory.
        # If it is set, then SaveEnv constructor will
        # load arguments from that trial record, will overwrite.
        # If not, then SaveEnv constructor will save configurations
        # as: ``<c.some_root_dir>/<trial_start_time>/config/config.json``

        save_env = SaveEnv(c)

    Args:
        merge_conf: Config to merge.
    """
    parser = argparse.ArgumentParser(description=__doc__)
    parser.add_argument("--conf", action="append")
    args = parser.parse_args()

    config_dict = {}
    if args.conf is not None:
        for env_str in args.conf:
            name, value = env_str.split('=')
            value = eval(value)
            config_dict[name] = value

    return merge_config((Config()
                         if merge_conf is None
                         else merge_conf), config_dict)


def load_config_file(json_file: str, merge_conf: Config = None) -> Config:
    """
    Get configs from a json file.

    Args:
        json_file: Path to the json config file.
        merge_conf: Config to merge.

    Return:
        configuration
    """
    # parse the configurations from the config json file provided
    with open(json_file, 'r') as config_file:
        config_dict = json.load(config_file)

    return merge_config((Config()
                         if merge_conf is None
                         else merge_conf), config_dict)


def save_config(conf: Config, json_file: str):
    """
    Dump config object to a json file.
    """
    with open(json_file, 'w') as config_file:
        json.dump(conf.data, config_file, sort_keys=True, indent=4)


def merge_config(conf: Config, merge: Union[dict, Config]) -> Config:
    """
    Merge config object with a dictionary, or a Config object,
    same keys in the ``conf`` will be overwritten by keys
    in ``merge``.
    """
    new_conf = copy.deepcopy(conf)
    if isinstance(merge, dict):
        for k, v in merge.items():
            new_conf[k] = v
    else:
        for k, v in merge.data.items():
            if k not in new_conf.const_attrs:
                new_conf[k] = v
    return new_conf



import traceback


class RemoteTraceback(Exception):  # pragma: no cover
    """
    Remote traceback, rebuilt by ``_rebuild_exc`` from pickled original
    traceback ExceptionWithTraceback, should be thrown on the master
    side.
    """
    def __init__(self, tb):
        self.tb = tb

    def __str__(self):
        return self.tb


def _rebuild_exc(exc, tb):
    exc.__cause__ = RemoteTraceback(tb)
    return exc


class ExceptionWithTraceback:  # pragma: no cover
    def __init__(self, exc: Exception, tb: str = None):
        """
        This exception class is used by slave processes to capture
        exceptions thrown during execution and send back throw queues
        to their master.

        Args:
            exc: Your exception.
            tb: An optional traceback, by default is is set to
                ``exc.__traceback__``
        """
        if tb is None:
            tb = exc.__traceback__
        tb = traceback.format_exception(type(exc), exc, tb)
        tb = ''.join(tb)
        self.exc = exc
        self.tb = '\n"""\n%s"""' % tb

    def __reduce__(self):
        # Used by pickler
        return _rebuild_exc, (self.exc, self.tb)





from .a2c import *
from machin.parallel.server import PushPullGradServer
from torch.optim import Adam


class A3C(A2C):
    """
    A3C framework.
    """
    def __init__(self,
                 actor: Union[NeuralNetworkModule, nn.Module],
                 critic: Union[NeuralNetworkModule, nn.Module],
                 criterion: Callable,
                 grad_server: Tuple[PushPullGradServer,
                                    PushPullGradServer],
                 *_,
                 entropy_weight: float = None,
                 value_weight: float = 0.5,
                 gradient_max: float = np.inf,
                 gae_lambda: float = 1.0,
                 discount: float = 0.99,
                 update_times: int = 50,
                 replay_size: int = 500000,
                 replay_device: Union[str, t.device] = "cpu",
                 replay_buffer: Buffer = None,
                 visualize: bool = False,
                 **__):
        """
        See Also:
            :class:`.A2C`

        Note:
            A3C algorithm relies on parameter servers to synchronize
            parameters of actor and critic models across samplers (
            interact with environment) and trainers (using samples
            to train.

            The parameter server type :class:`.PushPullGradServer`
            used here utilizes gradients calculated by trainers:

            1. perform a "sum" reduction process on the collected
            gradients, then apply this reduced gradient to the model
            managed by its primary reducer

            2. push the parameters of this updated managed model to
            a ordered key-value server so that all processes,
            including samplers and trainers, can access the updated
            parameters.

            The ``grad_servers`` argument is a pair of accessors to
            two :class:`.PushPullGradServerImpl` class.

        Args:
            actor: Actor network module.
            critic: Critic network module.
            optimizer: Optimizer used to optimize ``actor`` and ``critic``.
            criterion: Criterion used to evaluate the value loss.
            grad_server: Custom gradient sync server accessors, the first
                server accessor is for actor, and the second one is for critic.
            entropy_weight: Weight of entropy in your loss function, a positive
                entropy weight will minimize entropy, while a negative one will
                maximize entropy.
            value_weight: Weight of critic value loss.
            gradient_max: Maximum gradient.
            gae_lambda: :math:`\\lambda` used in generalized advantage
                estimation.
            discount: :math:`\\gamma` used in the bellman function.
            update_times: Number of update iterations per sample period. Buffer
                will be cleared after ``update()``
            replay_size: Replay buffer size. Not compatible with
                ``replay_buffer``.
            replay_device: Device where the replay buffer locates on, Not
                compatible with ``replay_buffer``.
            replay_buffer: Custom replay buffer.
            visualize: Whether visualize the network flow in the first pass.
        """
        # Adam is just a placeholder here, the actual optimizer is
        # set in parameter servers
        super(A3C, self).__init__(actor, critic, Adam, criterion,
                                  entropy_weight=entropy_weight,
                                  value_weight=value_weight,
                                  gradient_max=gradient_max,
                                  gae_lambda=gae_lambda,
                                  discount=discount,
                                  update_times=update_times,
                                  replay_size=replay_size,
                                  replay_device=replay_device,
                                  replay_buffer=replay_buffer,
                                  visualize=visualize)
        # disable local stepping
        self.actor_optim.step = lambda: None
        self.critic_optim.step = lambda: None
        self.actor_grad_server, self.critic_grad_server = \
            grad_server[0], grad_server[1]
        self.is_syncing = True

    def set_sync(self, is_syncing):
        self.is_syncing = is_syncing

    def manual_sync(self):
        self.actor_grad_server.pull(self.actor)
        self.critic_grad_server.pull(self.critic)

    def act(self, state: Dict[str, Any], **__):
        # DOC INHERITED
        if self.is_syncing:
            self.actor_grad_server.pull(self.actor)
        return super(A3C, self).act(state)

    def _eval_act(self,
                  state: Dict[str, Any],
                  action: Dict[str, Any],
                  **__):
        # DOC INHERITED
        if self.is_syncing:
            self.actor_grad_server.pull(self.actor)
        return super(A3C, self)._eval_act(state, action)

    def _criticize(self, state: Dict[str, Any], *_, **__):
        # DOC INHERITED
        if self.is_syncing:
            self.critic_grad_server.pull(self.critic)
        return super(A3C, self)._criticize(state)

    def update(self,
               update_value=True,
               update_policy=True,
               concatenate_samples=True,
               **__):
        # DOC INHERITED
        org_sync = self.is_syncing
        self.is_syncing = False
        super(A3C, self).update(update_value, update_policy,
                                concatenate_samples)
        self.is_syncing = org_sync
        self.actor_grad_server.push(self.actor)
        self.critic_grad_server.push(self.critic)



from typing import Dict, Iterable, Any
from os.path import join
from .logging import default_logger

import os
import re
import shutil
import torch as t
import torch.nn as nn


def prep_clear_dirs(dirs: Iterable[str]):
    """
    Args:
         dirs: a list of directories to clear
    """
    for dir_ in dirs:
        file_list = [f for f in os.listdir(dir_)]
        for f in file_list:
            f = os.path.join(dir_, f)
            if os.path.isfile(f) or os.path.islink(f):
                os.unlink(f)
            elif os.path.isdir(f):
                shutil.rmtree(f)


def prep_create_dirs(dirs: Iterable[str]):
    """
    Note: will recursively create directories.

    Args:
        dirs: a list of directories to create if these directories
            are not found.
    """
    for dir_ in dirs:
        if not os.path.exists(dir_):
            os.makedirs(dir_)


def prep_load_state_dict(model: nn.Module,
                         state_dict: Any):
    """
    Automatically load a **loaded state dictionary**

    Note:
        This function handles tensor device remapping.
    """
    for name, param in model.state_dict().items():
        state_dict[name].to(param.device)
    model.load_state_dict(state_dict)


def prep_load_model(model_dir: str,
                    model_map: Dict[str, t.nn.Module],
                    version: int = None,
                    quiet: bool = False,
                    logger: Any = None):
    """
    Automatically find and load models.

    Args:
        model_dir: Directory to save models.
        model_map: Model saving map.
        version: Version to load, if specified, otherwise automatically
            find the latest version.
        quiet: Raise no error if no valid version could be found.
        logger: Logger to use.
    """
    if not os.path.exists(model_dir) or not os.path.isdir(model_dir):
        raise RuntimeError("Model directory doesn't exist!")
    if logger is None:
        logger = default_logger

    version_map = {}
    for net_name in model_map.keys():
        version_map[net_name] = set()
    models = os.listdir(model_dir)
    for m in models:
        match = re.fullmatch("([a-zA-Z0-9_-]+)_([0-9]+)\\.pt$", m)
        if match is not None:
            n = match.group(1)
            v = int(match.group(2))
            if n in model_map:
                version_map[n].add(v)
    if version is not None:
        is_version_found = [version in version_map[name]
                            for name in model_map.keys()]
        if all(is_version_found):
            logger.info("Specified version found, using version: {}"
                        .format(version))
            for net_name, net in model_map.items():
                net = net  # type: nn.Module
                state_dict = t.load(join(
                    model_dir, "{}_{}.pt".format(net_name, version)),
                    map_location="cpu"
                ).state_dict()
                prep_load_state_dict(net, state_dict)
            return
        else:
            for ivf, net_name in zip(is_version_found, model_map.keys()):
                if not ivf:
                    logger.warning(
                        "Specified version {} for network {} is invalid"
                        .format(version, net_name)
                    )

    logger.info("Begin auto find")
    # use the valid, latest model
    common = set.intersection(*version_map.values())
    if len(common) == 0:
        if not quiet:
            raise RuntimeError("Cannot find a valid version for all models!")
        else:
            return
    version = max(common)
    logger.info("Using version: {}".format(version))
    for net_name, net in model_map.items():
        state_dict = t.load(join(
            model_dir, "{}_{}.pt".format(net_name, version)),
            map_location="cpu"
        ).state_dict()
        prep_load_state_dict(net, state_dict)



from .world import (
    World, CollectiveGroup, RpcGroup,
    get_world, get_cur_rank, get_cur_name
)

from . import world

__all__ = [
    "World", "CollectiveGroup", "RpcGroup",
    "get_world", "get_cur_rank", "get_cur_name",
    "world"
]



import os
import threading
import multiprocessing.pool as pool
from multiprocessing.pool import TERMINATE
from torch.multiprocessing.pool import clean_worker
from torch.multiprocessing import get_context

from .pickle import dumps, loads
from .queue import SimpleQueue, MultiP2PQueue


def proxy_caller(*input_):
    """
    Call a serialized function and return results.
    """
    if len(input_) == 1:
        func_str, args, kwargs = input_[0]
    else:
        func_str, args, kwargs = input_
    func = loads(func_str)
    return func(*args, **kwargs)


def proxy_ctx_caller(*input_):
    """
    Call a serialized function with worker context and return results.
    """
    if len(input_) == 1:
        func_str, args, kwargs = input_[0]
    else:
        func_str, args, kwargs = input_
    func = loads(func_str)
    return func(CtxPoolStorage.storage, *args, **kwargs)


def proxy_dumper(recurse, copy_tensor, func, args_list):
    """
    Serialize a function so it can be called.

    Returns:
        List[function string, arguments...]
    """
    # recurse will enable context variable saving
    dump = dumps(func, recurse=recurse, copy_tensor=copy_tensor)
    for args in args_list:
        yield [dump, args, {}]


class Pool(pool.Pool):
    """
    Enhanced multiprocessing pool for pytorch, provides:
     1. Support for lambdas and local functions.
     2. Ability to select the tensor serialize scheme.
    """

    def __init__(self, processes=None, initializer=None, initargs=(),
                 maxtasksperchild=None, is_recursive=False, is_daemon=True,
                 is_copy_tensor=True, share_method=None):
        """
        Note:
            To share "cpu" tensors in shared memory, you must set::

                is_copy_tensor=False,
                share_method="cpu"

            To share "cuda" tensors, you must set::

                is_copy_tensor=False,
                share_method="cuda"

        Note:
            The default context used in pool is "spawn", to avoid any issues
            brought by "fork". "fork" will only be used if you want to pass
            cpu tensors in shared memory.

        Args:
            processes: Number of processes in the pool.
            initializer: Initializer function executed by the pool/
            initargs: Args passed to the init function.
            maxtasksperchild: Maximum number of tasks per worker process.
            is_recursive: Set to ``True`` to support local functions
                and lambdas.
            is_daemon: Whether worker processes in the pool are started as
                daemon processes.
            is_copy_tensor: Whether to copy tensors or pass tensors by
                reference to worker processes.
            share_method: If ``is_copy_tensor`` is ``False``, you must
                specify this argument. "cpu" means you may use cpu tensors
                in the shared memory, "cuda" means cuda tensors, you can only
                specify one share method.
        """
        if processes is None:
            processes = os.cpu_count() or 1
        if processes < 1:
            raise ValueError("Number of processes must be at least 1")
        self._processes = processes

        self._is_recursive = is_recursive
        self._is_daemon = is_daemon
        self._is_copy_tensor = is_copy_tensor
        self._caller = proxy_caller

        context = get_context("spawn")
        if not is_copy_tensor:
            if share_method not in ("cpu", "cuda"):
                raise RuntimeError('Invalid share method: "{}"'
                                   .format(share_method))
            if share_method == "cpu":
                context = get_context("fork")

        super(Pool, self).__init__(
            processes=processes,
            initializer=initializer,
            initargs=initargs,
            maxtasksperchild=maxtasksperchild,
            context=context
        )

    def _setup_queues(self):
        # queues are only used to send dumped strings
        self._inqueue = SimpleQueue(ctx=self._ctx)
        self._outqueue = SimpleQueue(ctx=self._ctx)
        self._quick_put = self._inqueue.quick_put
        self._quick_get = self._outqueue.quick_get

    def apply(self, func, args=(), kwds=None):
        # DOC INHERITED
        if kwds is None:
            kwds = {}
        return pool.Pool.apply(self, self._caller,
                               [(dumps(func,
                                       recurse=self._is_recursive,
                                       copy_tensor=self._is_copy_tensor),
                                 args, kwds)])

    def apply_async(self, func, args=(), kwds=None, callback=None,
                    error_callback=None):
        # DOC INHERITED
        if kwds is None:
            kwds = {}
        return pool.Pool.apply_async(self, self._caller,
                                     [(dumps(func,
                                             recurse=self._is_recursive,
                                             copy_tensor=self._is_copy_tensor),
                                       args, kwds)])

    def map(self, func, iterable, chunksize=None):
        # DOC INHERITED
        return pool.Pool.map(self, self._caller,
                             proxy_dumper(
                                 self._is_recursive,
                                 self._is_copy_tensor,
                                 func,
                                 [(arg,) for arg in iterable]
                             ),
                             chunksize)

    def map_async(self, func, iterable, chunksize=None, callback=None,
                  error_callback=None):
        # DOC INHERITED
        return pool.Pool.map_async(self, self._caller,
                                   proxy_dumper(
                                       self._is_recursive,
                                       self._is_copy_tensor,
                                       func,
                                       [(arg,) for arg in iterable]
                                   ),
                                   chunksize, callback, error_callback)

    def imap(self, func, iterable, chunksize=1):
        # DOC INHERITED
        return pool.Pool.imap(self, self._caller,
                              proxy_dumper(
                                  self._is_recursive,
                                  self._is_copy_tensor,
                                  func,
                                  [(arg,) for arg in iterable]
                              ),
                              chunksize)

    def imap_unordered(self, func, iterable, chunksize=1):
        # DOC INHERITED
        return pool.Pool.imap_unordered(self, self._caller,
                                        proxy_dumper(
                                            self._is_recursive,
                                            self._is_copy_tensor,
                                            func,
                                            [(arg,) for arg in iterable]
                                        ),
                                        chunksize)

    def starmap(self, func, iterable, chunksize=None):
        # DOC INHERITED
        return pool.Pool.starmap(self, self._caller,
                                 proxy_dumper(
                                     self._is_recursive,
                                     self._is_copy_tensor,
                                     func,
                                     iterable
                                 ),
                                 chunksize)

    def starmap_async(self, func, iterable, chunksize=None, callback=None,
                      error_callback=None):
        # DOC INHERITED
        return pool.Pool.starmap_async(self, self._caller,
                                       proxy_dumper(
                                           self._is_recursive,
                                           self._is_copy_tensor,
                                           func,
                                           iterable
                                       ),
                                       chunksize, callback, error_callback)

    def _repopulate_pool(self):
        """
        Bring the number of pool processes up to the specified number,
        for use after reaping workers which have exited.
        """
        for _ in range(self._processes - len(self._pool)):
            # changed worker -> clean_worker
            args = (self._inqueue, self._outqueue,
                    self._initializer,
                    self._initargs, self._maxtasksperchild)
            if hasattr(self, '_wrap_exception'):
                args += (self._wrap_exception,)
            worker = self.Process(target=clean_worker, args=args)
            self._pool.append(worker)
            worker.name = worker.name.replace('Process', 'PoolWorker')
            worker.daemon = self._is_daemon
            worker.start()
            pool.util.debug('added worker')

    def size(self):
        """
        Returns:
            The number of workers in pool.
        """
        return self._processes

    def __reduce__(self):
        raise RuntimeError("Process pool is not reducible.")


class P2PPool(Pool):
    def _setup_queues(self):
        # queues are only used to send dumped strings
        self._inqueue = MultiP2PQueue(self._processes)
        self._outqueue = MultiP2PQueue(self._processes)
        self._quick_put = self._inqueue.put
        self._quick_get = self._outqueue.get

    def _repopulate_pool(self):
        """
        Bring the number of pool processes up to the specified number,
        for use after reaping workers which have exited.
        """
        for idx in range(self._processes - len(self._pool)):
            # changed worker -> clean_worker
            args = (self._inqueue.get_sub_queue(idx),
                    self._outqueue.get_sub_queue(idx),
                    self._initializer,
                    self._initargs, self._maxtasksperchild)
            if hasattr(self, '_wrap_exception'):
                args += (self._wrap_exception,)
            worker = self.Process(target=clean_worker, args=args)
            self._pool.append(worker)
            worker.name = worker.name.replace('Process', 'PoolWorker')
            worker.daemon = self._is_daemon
            worker.start()
            pool.util.debug('added worker')

    def close(self):
        # we cannot rely on sentinels to shutdown worker processes
        # since there is no gaurantee that each worker will get 1
        # "None" sentinel
        self.terminate()

    @classmethod
    def _terminate_pool(cls, taskqueue, inqueue, outqueue, pool,
                        worker_handler, task_handler, result_handler, cache):

        worker_handler._state = TERMINATE
        task_handler._state = TERMINATE
        result_handler._state = TERMINATE

        # send sentinels so that handlers will exit
        outqueue.put(None)

        if threading.current_thread() is not worker_handler:
            worker_handler.join()

        if threading.current_thread() is not task_handler:
            task_handler.join()

        if threading.current_thread() is not result_handler:
            result_handler.join()

        # terminate workers directly
        if pool and hasattr(pool[0], 'terminate'):
            for p in pool:
                p.terminate()
                p.join(timeout=1e-1)


class CtxPoolStorage:
    """
    This storage class is used by all :class:`.CtxPool` instances.
    However, since for each worker process, they have different
    memory spaces, ``storage`` is unique for all workers.

    ``storage`` is accessed on the client process side.
    """
    storage = None


class CtxPool(Pool):
    """
    Pool with context for each worker. your function must accept a ``ctx``
    object as your first non-keyword argument.

    If ``worker_contexts`` is not specified, then ``ctx`` will be ``None``.

    The length of ``worker_contexts`` must be the same as ``processes``
    """

    def __init__(self, processes: int, initializer=None, initargs=(),
                 maxtasksperchild=None, worker_contexts=None,
                 is_recursive=False, is_daemon=True,
                 is_copy_tensor=True, share_method=None):

        if worker_contexts is not None:
            if len(worker_contexts) != processes:
                raise ValueError("Context number is not equal to the number of "
                                 "pool workers.")
        else:
            worker_contexts = [None] * processes

        super(CtxPool, self).__init__(
            processes=processes,
            initializer=self._init_with_context,
            initargs=(worker_contexts, initializer) + initargs,
            maxtasksperchild=maxtasksperchild,
            is_recursive=is_recursive,
            is_daemon=is_daemon,
            is_copy_tensor=is_copy_tensor,
            share_method=share_method
        )
        self._caller = proxy_ctx_caller

    def _repopulate_pool(self):
        """
        Bring the number of pool processes up to the specified number,
        for use after reaping workers which have exited.
        """
        # Get existing process ids:
        ids = {p.id for p in self._pool}
        need_ids = set(range(self._processes)) - ids

        for _, id in zip(range(self._processes - len(self._pool)), need_ids):
            initargs = list(self._initargs)

            # Unpack context
            initargs[0] = initargs[0][id]

            args = (self._inqueue, self._outqueue,
                    self._initializer,
                    initargs, self._maxtasksperchild)

            if hasattr(self, '_wrap_exception'):
                args += (self._wrap_exception,)

            # changed worker -> clean_worker
            worker = self.Process(target=clean_worker, args=args)
            worker.id = id
            self._pool.append(worker)
            worker.name = worker.name.replace('Process', 'CtxPoolWorker')
            worker.daemon = self._is_daemon
            worker.start()
            pool.util.debug('added worker')

    @staticmethod
    def _init_with_context(context, init_func, *initargs):
        CtxPoolStorage.storage = context
        if init_func is not None:
            init_func(*initargs)


class ThreadPool(pool.ThreadPool):
    """
    A typical thread pool.
    """

    # Multiprocessing pool is badly written.
    # python IDEs will complain a lot.

    # Seems that manually adding gc
    # (when using torch.multiprocessing.pool.clean_worker as worker function)
    # will cause thread-pool to hang on this function on exit:
    # _wait_for_tstate_lock()

    # so _repopulate_pool is not overloaded

    def size(self):
        """
        Returns:
            The number of workers in pool.
        """
        return len(self._pool)

    def __reduce__(self):
        raise RuntimeError("Thread pool is not reducible.")


class CtxThreadPool(ThreadPool):
    _context = threading.local()

    def __init__(self, processes: int, initializer=None, initargs=(),
                 worker_contexts=None):
        if worker_contexts is not None:
            if len(worker_contexts) != processes:
                raise ValueError("Context number is not equal to the number of "
                                 "pool workers.")
        else:
            worker_contexts = [None] * processes

        super(CtxThreadPool, self).__init__(
            processes=processes,
            initializer=self._init_with_context,
            initargs=(worker_contexts, initializer) + initargs
        )

    def apply(self, func, args=(), kwds=None):
        if kwds is None:
            kwds = {}
        return super(CtxThreadPool, self).apply_async(
            self._wrap_func(func), args, kwds
        ).get()

    def apply_async(self, func, args=(), kwds=None, callback=None,
                    error_callback=None):
        if kwds is None:
            kwds = {}
        return super(CtxThreadPool, self).apply_async(
            self._wrap_func(func), args, kwds, callback, error_callback
        )

    def map(self, func, iterable, chunksize=None):
        return super(CtxThreadPool, self).map(
            self._wrap_func(func), iterable, chunksize
        )

    def map_async(self, func, iterable, chunksize=None, callback=None,
                  error_callback=None):
        return super(CtxThreadPool, self).map_async(
            self._wrap_func(func), iterable, chunksize,
            callback, error_callback
        )

    def imap(self, func, iterable, chunksize=1):
        return super(CtxThreadPool, self).imap(
            self._wrap_func(func), iterable, chunksize
        )

    def imap_unordered(self, func, iterable, chunksize=1):
        return super(CtxThreadPool, self).imap_unordered(
            self._wrap_func(func), iterable, chunksize
        )

    def starmap(self, func, iterable, chunksize=None):
        return super(CtxThreadPool, self).starmap(
            self._wrap_func(func), iterable, chunksize
        )

    def starmap_async(self, func, iterable, chunksize=None, callback=None,
                      error_callback=None):
        return super(CtxThreadPool, self).starmap_async(
            self._wrap_func(func), iterable, chunksize,
            callback, error_callback
        )

    def _repopulate_pool(self):
        """
        Bring the number of pool processes up to the specified number,
        for use after reaping workers which have exited.
        """
        # Get existing process ids:
        ids = {p.id for p in self._pool}
        need_ids = set(range(self._processes)) - ids

        for _, id in zip(range(self._processes - len(self._pool)), need_ids):
            initargs = list(self._initargs)

            # Unpack context
            initargs[0] = initargs[0][id]

            args = (self._inqueue, self._outqueue,
                    self._initializer,
                    initargs, self._maxtasksperchild)

            if hasattr(self, '_wrap_exception'):
                args += (self._wrap_exception,)

            # changed worker -> clean_worker
            worker = self.Process(target=pool.worker, args=args)
            worker.id = id
            self._pool.append(worker)
            worker.name = worker.name.replace('Process', 'CtxThreadPoolWorker')
            worker.start()
            pool.util.debug('added worker')

    @classmethod
    def _wrap_func(cls, func):
        def call(*args, **kwargs):
            ctx = cls._context.storage
            return func(ctx, *args, **kwargs)

        return call

    @staticmethod
    def _init_with_context(context, init_func, *initargs):
        CtxThreadPool._context.storage = context
        if init_func is not None:
            init_func(*initargs)



from machin.utils.logging import default_logger
from machin.model.nets.base import static_module_wrapper
import inspect
import torch
import torch.nn as nn


def soft_update(target_net: nn.Module,
                source_net: nn.Module,
                update_rate) -> None:
    """
    Soft update target network's parameters.

    Args:
        target_net: Target network to be updated.
        source_net: Source network providing new parameters.
        update_rate: Update rate.

    Returns:
        None
    """
    with torch.no_grad():
        for target_param, param in zip(target_net.parameters(),
                                       source_net.parameters()):
            target_param.data.copy_(
                target_param.data * (1.0 - update_rate) +
                param.data.to(target_param.device) * update_rate
            )


def hard_update(target_net: nn.Module,
                source_net: nn.Module) -> None:
    """
    Hard update (directly copy) target network's parameters.

    Args:
        target_net: Target network to be updated.
        source_net: Source network providing new parameters.
    """

    for target_buffer, buffer in zip(target_net.buffers(),
                                     source_net.buffers()):
        target_buffer.data.copy_(buffer.data)
    for target_param, param in zip(target_net.parameters(),
                                   source_net.parameters()):
        target_param.data.copy_(param.data)


def determine_device(model):
    devices = set()
    for k, v in model.named_parameters():
        devices.add(str(v.device))
    return list(devices)


def safe_call(model, *named_args):
    """
    Call a model and discard unnecessary arguments. safe_call will automatically
    move tensors in named_args to the input device of the model

    Any input tensor in named_args must not be contained inside any container,
    such as list, dict, tuple, etc. Because they will be automatically moved
    to the input device of the specified model.

    Args:
        model: Model to be called, must be a wrapped nn.Module or an instance of
               NeuralNetworkModule.
        named_args: A dictionary of argument, key is argument's name, value is
                    argument's value.

    Returns:
        Whatever returned by your module. If result is not a tuple, always
        wrap results inside a tuple
    """
    org_model = None
    if isinstance(model,
                  (nn.parallel.DistributedDataParallel,
                   nn.parallel.DataParallel)):
        org_model = model
        model = model.module
    if (not hasattr(model, "input_device") or
            not hasattr(model, "output_device")):
        # try to automatically determine the input & output device of the model
        model_type = type(model)
        device = determine_device(model)
        if len(device) > 1:
            raise RuntimeError("Failed to automatically determine i/o device "
                               "of your model: {}\n"
                               "Detected multiple devices: {}\n"
                               "You need to manually specify i/o device of "
                               "your model.\n"
                               "Wrap your model of type nn.Module with one "
                               "of: \n"
                               "1. static_module_wrapper "
                               "from machin.model.nets.base \n"
                               "1. dynamic_module_wrapper "
                               "from machin.model.nets.base \n"
                               "Or construct your own module & model with: \n"
                               "NeuralNetworkModule from machin.model.nets.base"
                               .format(model_type, device))
        else:
            # assume that i/o devices are the same as parameter device
            # print a warning
            default_logger.warning("You have not specified the i/o device of "
                                   "your model {}, automatically determined and"
                                   " set to: {}\n"
                                   "The framework is not responsible for any "
                                   "un-matching device issues caused by this "
                                   "operation.".format(model_type, device[0]))
            model = static_module_wrapper(model, device[0], device[0])

    input_device = model.input_device
    arg_spec = inspect.getfullargspec(model.forward)
    # exclude self in arg_spec.args
    args = arg_spec.args[1:] + arg_spec.kwonlyargs
    if arg_spec.defaults is not None:
        args_with_defaults = args[-len(arg_spec.defaults):]
    else:
        args_with_defaults = []
    required_args = (set(args) - set(args_with_defaults) -
                     set(arg_spec.kwonlydefaults.keys()
                         if arg_spec.kwonlydefaults is not None
                         else []))
    args_dict = {}

    # fill in args
    for na in named_args:
        for k, v in na.items():
            if k in args:
                if torch.is_tensor(v):
                    args_dict[k] = v.to(input_device)
                else:
                    args_dict[k] = v

    # check for necessary args
    missing = required_args - set(args_dict.keys())
    if len(missing) > 0:
        raise RuntimeError("\n"
                           "The signature of the forward function of Model {} "
                           "is {}\n"
                           "Missing required arguments: {}, "
                           "check your storage functions."
                           .format(type(model), required_args, missing))

    if org_model is not None:
        result = org_model(**args_dict)
    else:
        result = model(**args_dict)

    if isinstance(result, tuple):
        return result
    else:
        return (result,)


def safe_return(result):
    if len(result) == 1:
        return result[0]
    else:
        return result


asserted_output_is_probs = False


def assert_output_is_probs(tensor):
    global asserted_output_is_probs
    if asserted_output_is_probs:
        return
    asserted_output_is_probs = True

    if tensor.dim() == 2 and \
            torch.all(torch.abs(torch.sum(tensor, dim=1) - 1.0) < 1e-5) and \
            torch.all(tensor > 0):
        return
    else:
        raise ValueError("Input tensor is not a probability tensor, it must "
                         "have 2 dimensions (0 and 1), a sum of 1.0 for each "
                         "row in dimension 1, and a positive value for each "
                         "element.")



from typing import Any
from dill import Pickler as DillPickler, loads as d_loads
from torch.multiprocessing import set_sharing_strategy
from torch.multiprocessing.reductions import (
    reduce_event, reduce_storage, reduce_tensor
)
import io
import copyreg
import torch as t

# strategy "file_descriptor" will not work if sender
# process has been terminated before receiver process receives the tensor
# because the receiver needs to connect to sender to get FDs
set_sharing_strategy("file_system")


def mark_static_module(module: Any):  # pragma: no cover
    """
    Some modules are **static**, which means they are stateless
    and will remain the same whether you import it in process A
    or process B.

    If your module contains reference to functions, objects
    or anything inside a CDLL (usually the reference is a
    pointer), it is not picklable by dill, and will cause
    nasty errors, however, by marking this module as "Static",
    dill will recognize this module as a builtin module and
    not saving the states of this module, dill will only save
    a reference to it in this situation.

    Args:
        module: Some module which imports CDLLs by hand and
            not using pybind11.
    """
    del module.__file__


def _rebuild_full(data):
    buffer = io.BytesIO(data)
    return t.load(buffer)


def _reduce_full(obj):
    # supports saving tensors, storage, etc.
    # will always save all data and not by reference.
    buffer = io.BytesIO()
    t.save(obj, buffer)
    return _rebuild_full, (buffer.getvalue(),)


class Pickler(DillPickler):
    """
    Note:
        Picklers shares ".dispatch" among instances, and owns
        "dispatch_table" per instance.

        The base Pickler (not dill, from builtin pickle library),
        will first look up the default dump method in ".dispatch", if
        no valid method is found, it will try to find a custom dump
        method in ".dispatch_table".
    """
    def __init__(self, file, recurse=False, copy_tensor=False):
        super(Pickler, self).__init__(
            file, byref=False, recurse=recurse
        )
        self.dispatch_table = copyreg.dispatch_table.copy()
        if not copy_tensor:
            # register the reduction methods provided by pytorch
            # same as init_reductions() in
            # torch.multiprocessing.reductions

            # In this case, receiver processes must be created by "fork",
            # and _share_memory()/share_memory() must be invoked on all
            # tensors/modules.
            # Otherwise "cpu" tensors will probably get a serious exception,
            # because the receiver processes are only getting pointers.
            # "cuda" tensors should will be fine
            self.dispatch_table[t.cuda.Event] = reduce_event
            for typ in t._storage_classes:
                self.dispatch_table[typ] = reduce_storage
            for typ in t._tensor_classes:
                self.dispatch_table[typ] = reduce_tensor
            self.dispatch_table[t.Tensor] = reduce_tensor
            self.dispatch_table[t.nn.parameter.Parameter] = reduce_tensor

        else:
            self.dispatch_table[t.cuda.Event] = reduce_event
            for typ in t._storage_classes:
                self.dispatch_table[typ] = _reduce_full
            for typ in t._tensor_classes:
                self.dispatch_table[typ] = _reduce_full
            self.dispatch_table[t.Tensor] = _reduce_full
            self.dispatch_table[t.nn.parameter.Parameter] = _reduce_full


def dumps(obj, recurse=False, copy_tensor=True):
    """
    Convert objects to bytes. Works for cpu and gpu tensors.

    Warning:
        Till pytorch 1.5.0, there is a bug for referenced gpu tensors,
        which would require users to keep shared gpu tensors during
        the whole process life and not reassigning / deleting them,
        however, you may refill them with different values.

        See `here <https://github.com/pytorch/pytorch/issues/39541>`_

    Args:
        obj: Object to dump.
        recurse: Enable recursive dumping, enable this to dump local
            functions and lambdas.
        copy_tensor: Whether to dump tensors, storage as a full copy.
            If it is set to "False", then dumped tensors must either
            locate on GPUs or in shared memory.

    Returns:
        Bytes.
    """
    buffer = io.BytesIO()
    pickler = Pickler(buffer, recurse, copy_tensor)
    pickler.dump(obj)
    return buffer.getvalue()


loads = d_loads



from typing import Union, Dict, List, Tuple, Callable, Any
import numpy as np
import torch as t
import torch.nn as nn

from machin.frame.buffers.buffer_d import Transition, DistributedBuffer
from machin.model.nets.base import NeuralNetworkModule
from .base import TorchFramework
from .utils import safe_call

from machin.parallel.server import PushPullModelServer
from machin.parallel.distributed import RpcGroup


def _make_tensor_from_batch(batch: List[Any], device, concatenate):
    """
    Used to convert compact every attribute of every step of a whole episode
    into a single tensor.

    Args:
        batch: A list of tensor or scalar. If elements are tensors, they will
            be concatenated in dimension 0.
        device: Device to move tensors to.
        concatenate: Whether to perform concatenation or not, only True
            for major and sub attributes in ``Transition``.

    Returns:
        A tensor if ``concatenate`` is ``True``, otherwise the original List.
    """
    if len(batch) == 0:
        return None
    if concatenate:
        item = batch[0]
        batch_size = len(batch)
        if t.is_tensor(item):
            return t.cat([it.to(device) for it in batch], dim=0).to(device)
        else:
            return t.tensor(batch, device=device).view(batch_size, -1)
    else:
        return batch


class EpisodeTransition(Transition):
    """
    A transition class which allows storing the whole episode as a
    single transition object, the batch dimension will be used to
    stack all transition steps.
    """
    def _check_validity(self):
        """
        Disable checking for batch size in the base :class:`.Transition`
        """
        super(Transition, self)._check_validity()


class EpisodeDistributedBuffer(DistributedBuffer):
    """
    A distributed buffer which stores each episode as a transition
    object inside the buffer.
    """
    def append(self, transition: Dict,
               required_attrs=("state", "action", "next_state",
                               "reward", "terminal", "action_log_prob")):
        transition = EpisodeTransition(**transition)
        super(EpisodeDistributedBuffer, self)\
            .append(transition, required_attrs=required_attrs)


class IMPALA(TorchFramework):
    """
    Massively parallel IMPALA framework.
    """

    _is_top = ["actor", "critic"]
    _is_restorable = ["actor", "critic"]

    def __init__(self,
                 actor: Union[NeuralNetworkModule, nn.Module],
                 critic: Union[NeuralNetworkModule, nn.Module],
                 optimizer: Callable,
                 criterion: Callable,
                 impala_group: RpcGroup,
                 model_server: Tuple[PushPullModelServer],
                 *_,
                 lr_scheduler: Callable = None,
                 lr_scheduler_args: Tuple[Tuple, Tuple] = (),
                 lr_scheduler_kwargs: Tuple[Dict, Dict] = (),
                 batch_size: int = 5,
                 learning_rate=0.001,
                 isw_clip_c: float = 1.0,
                 isw_clip_rho: float = 1.0,
                 entropy_weight: float = None,
                 value_weight: float = 0.5,
                 gradient_max: float = np.inf,
                 discount: float = 0.99,
                 replay_size: int = 500,
                 visualize: bool = False,
                 **__):
        """
        Note:
            Please make sure isw_clip_rho >= isw_clip_c
        Args:
            actor: Actor network module.
            critic: Critic network module.
            optimizer: Optimizer used to optimize ``actor`` and ``critic``.
            criterion: Criterion used to evaluate the value loss.
            impala_group: Group of all processes using the IMPALA framework,
                including all samplers and trainers.
            model_server: Custom model sync server accessor for ``actor``.
            lr_scheduler: Learning rate scheduler of ``optimizer``.
            lr_scheduler_args: Arguments of the learning rate scheduler.
            lr_scheduler_kwargs: Keyword arguments of the learning
                rate scheduler.
            batch_size: Batch size used during training.
            learning_rate: Learning rate of the optimizer, not compatible with
                ``lr_scheduler``.
            isw_clip_c: :math:`c` used in importance weight clipping.
            isw_clip_rho:
            entropy_weight: Weight of entropy in your loss function, a positive
                entropy weight will minimize entropy, while a negative one will
                maximize entropy.
            value_weight: Weight of critic value loss.
            gradient_max: Maximum gradient.
            discount: :math:`\\gamma` used in the bellman function.
            replay_size: Size of the local replay buffer.
            visualize: Whether visualize the network flow in the first pass.
        """
        self.batch_size = batch_size
        self.discount = discount
        self.value_weight = value_weight
        self.entropy_weight = entropy_weight
        self.grad_max = gradient_max
        self.isw_clip_c = isw_clip_c
        self.isw_clip_rho = isw_clip_rho
        self.visualize = visualize

        self.impala_group = impala_group

        self.actor = actor
        self.critic = critic
        self.actor_optim = optimizer(self.actor.parameters(),
                                     lr=learning_rate)
        self.critic_optim = optimizer(self.critic.parameters(),
                                      lr=learning_rate)
        self.replay_buffer = EpisodeDistributedBuffer(
            buffer_name="buffer", group=impala_group,
            buffer_size=replay_size
        )
        self.is_syncing = True
        self.actor_model_server = model_server[0]

        if lr_scheduler is not None:
            self.actor_lr_sch = lr_scheduler(
                self.actor_optim,
                *lr_scheduler_args[0],
                **lr_scheduler_kwargs[0],
            )
            self.critic_lr_sch = lr_scheduler(
                self.critic_optim,
                *lr_scheduler_args[1],
                **lr_scheduler_kwargs[1]
            )

        self.criterion = criterion

        super(IMPALA, self).__init__()

    def set_sync(self, is_syncing):
        self.is_syncing = is_syncing

    def manual_sync(self):
        self.actor_model_server.pull(self.actor)

    def act(self, state: Dict[str, Any], *_, **__):
        """
        Use actor network to give a policy to the current state.

        Returns:
            Anything produced by actor.
        """
        if self.is_syncing:
            self.actor_model_server.pull(self.actor)
        return safe_call(self.actor, state)

    def _eval_act(self,
                  state: Dict[str, Any],
                  action: Dict[str, Any],
                  *_, **__):
        """
        Use actor network to evaluate the log-likelihood of a given
        action in the current state.

        Returns:
            Anything produced by actor.
        """
        return safe_call(self.actor, state, action)

    def _criticize(self, state: Dict[str, Any], *_, **__):
        """
        Use critic network to evaluate current value.

        Returns:
            Value of shape ``[batch_size, 1]``
        """
        return safe_call(self.critic, state)[0]

    def store_transition(self, transition: Union[Transition, Dict]):
        """
        Warning:
            Not supported in IMPALA due to v-trace requirements.
        """
        raise NotImplementedError

    def store_episode(self, episode: List[Union[Transition, Dict]]):
        """
        Add a full episode of transition samples to the replay buffer.
        """
        if not isinstance(episode[0], Transition):
            episode = [Transition(**trans) for trans in episode]

        cc_episode = {}
        # In order to compute v-trace, we must reshape the whole
        # episode to make it look like a single Transition, because
        # v-trace need to see all future rewards.

        # therefore, only one entry will be stored into the buffer
        # each entry in the buffer is of shape [episode_length, ...]

        # In other frameworks. each entry in the buffer is of shape
        # [1, ...]
        for k, v in episode[0].items():
            if k in ("state", "action", "next_state"):
                tmp_dict = {}
                for sub_k in v.keys():
                    tmp_dict[sub_k] = _make_tensor_from_batch(
                        [item[k][sub_k] for item in episode],
                        self.replay_buffer.buffer_device, True
                    )
                cc_episode[k] = tmp_dict
            elif k in ("reward", "terminal", "action_log_prob"):
                cc_episode[k] = _make_tensor_from_batch(
                    [item[k] for item in episode],
                    self.replay_buffer.buffer_device, True
                )
            else:
                # currently, additional attributes are not supported.
                pass

        self.replay_buffer.append(cc_episode, required_attrs=(
            "state", "action", "next_state", "reward",
            "action_log_prob", "terminal"
        ))

    def update(self,
               update_value=True,
               update_policy=True,
               **__):
        """
        Update network weights by sampling from replay buffer.

        Note:
            Will always concatenate samples.

        Args:
            update_value: Whether to update the Q network.
            update_policy: Whether to update the actor network.

        Returns:
            mean value of estimated policy value, value loss
        """
        # sample a batch

        # Note: each episode is stored as a single sample entry,
        # the second dimension of all attributes is the length of episode,
        # the first dimension is always 1.

        # `batch_size` here means the number of episodes sampled, not
        # the number of steps sampled.

        # `concatenate` is False, because the length of each episode
        # might be different.

        batch_size, (state, action, reward, next_state,
                     terminal, action_log_prob) = \
            self.replay_buffer.sample_batch(self.batch_size,
                                            concatenate=False,
                                            device="cpu",
                                            sample_attrs=[
                                                "state", "action", "reward",
                                                "next_state", "terminal",
                                                "action_log_prob"],
                                            additional_concat_attrs=[
                                                "action_log_prob"
                                            ])
        # `state`, `action` and `next_state` should be dicts like:
        # {"attr1": [Tensor(ep1_length, ...),
        #            Tensor(ep2_length, ...)]}

        # `terminal`, `reward`, `action_log_prob` should be lists like:
        # [Tensor(ep1_length, 1), (ep2_length, 1)]

        # chain steps of all episodes together, make them look like:
        # ep1_step1, ep1_step2, ..., ep1_stepN, ep2_step1, ep2_step2 ...

        # store the length of each episode, so that we can find boundaries
        # between two episodes inside the chained "sample"
        all_length = [tensor.shape[0] for tensor in terminal]
        sum_length = sum(all_length)

        for major_attr in (state, action, next_state):
            for k, v in major_attr.items():
                major_attr[k] = t.cat(v, dim=0)
                assert major_attr[k].shape[0] == sum_length

        terminal = t.cat(terminal, dim=0).view(sum_length, 1)
        reward = t.cat(reward, dim=0).view(sum_length, 1)
        action_log_prob = t.cat(action_log_prob, dim=0).view(sum_length, 1)

        # Below are the v-trace process

        # Calculate c and rho first, because there is no dependency
        # between vector elements.
        _, cur_action_log_prob, entropy, *__ = self._eval_act(state, action)
        cur_action_log_prob = cur_action_log_prob.view(sum_length, 1)
        entropy = entropy.view(sum_length, 1)

        cur_action_log_prob = cur_action_log_prob.to("cpu")
        # similarity = pi(a_t|x_t)/mu(a_t|x_t)
        sim = t.exp(cur_action_log_prob - action_log_prob)
        c = t.min(t.full(sim.shape, self.isw_clip_c), sim)
        rho = t.min(t.full(sim.shape, self.isw_clip_rho), sim)

        # calculate delta V
        # delta_t V = rho_t(r_t + gamma * V(x_{t+1}) - V(x_t))
        # boundary elements (i.e, ep1_stepN) will have V(x_{t+1}) = 0
        value = self._criticize(state).view(sum_length, 1)
        next_value = self._criticize(next_state).view(sum_length, 1)
        next_value[terminal] = 0
        delta_v = rho * (reward +
                         self.discount * next_value.to("cpu")
                         - value.to("cpu"))

        # calculate v_s

        # vss is v_s shifted left by 1 element, i.e. becomes v_{s+1}
        # boundary elements (i.e, ep1_stepN) will have v_{s+1} = 0

        # do reversed cumulative product for each episode segment
        with t.no_grad():
            vs = t.zeros(value.shape)
            vss = t.zeros(value.shape)
            offset = 0
            for ep_len in all_length:
                # the last v_s of each episode should be 0
                # or V_t - rho_t * (r_t - V_t)? (since Vt+1 = 0)
                # Implementations such as
                # https://github.com/junjungoal/IMPALA-pytorch/blob/master
                # /agents/learner.py   use the first case, 0
                # 0 works well when rho=c=1 or 1 > rho >= c
                for rev_step in reversed(range(ep_len - 1)):
                    idx = offset + rev_step
                    vs[idx] = (value[idx] + delta_v[idx] +
                               self.discount * c[idx] *
                               (vs[idx + 1] - value[idx + 1]))
                # shift v_s to get v_{s+1}
                vss[offset: offset + ep_len - 1] = \
                    vs[offset + 1: offset + ep_len]

                # update offset
                offset += ep_len

        act_policy_loss = -(rho.detach() * cur_action_log_prob *
                            (reward + self.discount * vss - value).detach())
        if self.entropy_weight is not None:
            act_policy_loss += self.entropy_weight * entropy
        act_policy_loss = act_policy_loss.sum()

        value_loss = self.criterion(value, vs.to(value.device))

        # Update actor network
        if update_policy:
            self.actor.zero_grad()
            act_policy_loss.backward()
            nn.utils.clip_grad_norm_(
                self.actor.parameters(), self.grad_max
            )
            self.actor_optim.step()

        # Update critic network
        if update_value:
            self.critic.zero_grad()
            value_loss.backward()
            nn.utils.clip_grad_norm_(
                self.critic.parameters(), self.grad_max
            )
            self.critic_optim.step()

        # push actor model for samplers
        if isinstance(self.actor,
                      (nn.parallel.DataParallel,
                       nn.parallel.DistributedDataParallel)):
            self.actor_model_server.push(self.actor.module,
                                         pull_on_fail=False)
        else:
            self.actor_model_server.push(self.actor)
        return -act_policy_loss.item(), value_loss.item()

    def update_lr_scheduler(self):
        """
        Update learning rate schedulers.
        """
        if hasattr(self, "actor_lr_sch"):
            self.actor_lr_sch.step()
        if hasattr(self, "critic_lr_sch"):
            self.critic_lr_sch.step()



from typing import Union, Dict, List, Any, Callable
from threading import RLock
from ..transition import Transition
from .buffer import Buffer
from machin.parallel.distributed import RpcGroup
import torch as t
import numpy as np
import itertools as it


def _round_up(num):
    return int(np.ceil(num))


class DistributedBuffer(Buffer):
    def __init__(self, buffer_name: str, group: RpcGroup, buffer_size: int,
                 *_, **__):
        """
        Create a distributed replay buffer instance.

        To avoid issues caused by tensor device difference, all transition
        objects are stored in device "cpu".

        Distributed replay buffer constitutes of many local buffers held per
        process, transmissions between processes only happen during sampling.

        During sampling, the tensors in "state", "action" and "next_state"
        dictionaries, along with "reward", will be concatenated in dimension 0.
        any other custom keys specified in ``**kwargs`` will not be
        concatenated.

        .. seealso:: :class:`.Buffer`

        Note:
            Since ``append()`` operates on the local buffer, in order to
            append to the distributed buffer correctly, please make sure
            that your actor is also the local buffer holder, i.e. a member
            of the ``group``

        Args:
            buffer_size: Maximum local buffer size.
            group: Process group which holds this buffer.
            buffer_name: A unique name of your buffer.
        """
        super(DistributedBuffer, self).__init__(buffer_size, "cpu")
        self.buffer_name = buffer_name
        self.group = group

        assert group.is_member()

        # register services, so that we may access other buffers
        _name = "/" + group.get_cur_name()
        self.group.register(buffer_name + _name + "/_size_service",
                            self._size_service)
        self.group.register(buffer_name + _name + "/_clear_service",
                            self._clear_service)
        self.group.register(buffer_name + _name + "/_sample_service",
                            self._sample_service)
        self.wr_lock = RLock()

    def append(self, transition: Union[Transition, Dict],
               required_attrs=("state", "action", "next_state",
                               "reward", "terminal")):
        # DOC INHERITED
        with self.wr_lock:
            super(DistributedBuffer, self).append(
                transition, required_attrs=required_attrs)

    def clear(self):
        """
        Clear current local buffer.
        """
        with self.wr_lock:
            return super(DistributedBuffer, self).clear()

    def all_clear(self):
        """
        Remove all entries from all local buffers.
        """
        future = [
            self.group.registered_async(
                self.buffer_name + "/" + m + "/_clear_service"
            )
            for m in self.group.get_group_members()
        ]
        for fut in future:
            fut.wait()

    def size(self):
        """
        Returns:
            Length of current local buffer.
        """
        with self.wr_lock:
            return super(DistributedBuffer, self).size()

    def all_size(self):
        """
        Returns:
            Total length of all buffers.
        """
        future = []
        count = 0
        for m in self.group.get_group_members():
            future.append(self.group.registered_async(
                self.buffer_name + "/" + m + "/_size_service"
            ))
        for fut in future:
            count += fut.wait()
        return count

    def sample_batch(self,
                     batch_size: int,
                     concatenate: bool = True,
                     device: Union[str, t.device] = None,
                     sample_method: Union[Callable, str] = "random_unique",
                     sample_attrs: List[str] = None,
                     additional_concat_attrs: List[str] = None,
                     *_, **__) -> Any:
        # DOC INHERITED
        p_num = self.group.size()
        local_batch_size = _round_up(batch_size / p_num)

        future = [
            self.group.registered_async(
                self.buffer_name + "/" + m + "/_sample_service",
                args=(local_batch_size, sample_method)
            )
            for m in self.group.get_group_members()
        ]

        results = [fut.wait() for fut in future]
        all_batch_size = sum([r[0] for r in results])
        all_batch = list(it.chain(*[r[1] for r in results]))

        if device is None:
            device = "cpu"
        if sample_attrs is None:
            sample_attrs = all_batch[0].keys()
        if additional_concat_attrs is None:
            additional_concat_attrs = []

        return \
            all_batch_size, \
            Buffer.post_process_batch(all_batch, device, concatenate,
                                      sample_attrs, additional_concat_attrs)

    def _size_service(self):  # pragma: no cover
        return self.size()

    def _clear_service(self):  # pragma: no cover
        self.clear()

    def _sample_service(self, batch_size, sample_method):  # pragma: no cover
        if isinstance(sample_method, str):
            if not hasattr(self, "sample_method_" + sample_method):
                raise RuntimeError("Cannot find specified sample method: {}"
                                   .format(sample_method))
            sample_method = getattr(self, "sample_method_" + sample_method)

        # sample raw local batch from local buffer
        with self.wr_lock:
            local_batch_size, local_batch = sample_method(self.buffer,
                                                          batch_size)

        return local_batch_size, local_batch



from machin.frame.buffers.prioritized_buffer import PrioritizedBuffer
from machin.utils.logging import default_logger
# pylint: disable=wildcard-import, unused-wildcard-import
from .dqn import *


class DQNPer(DQN):
    """
    DQN with prioritized replay. It is based on Double DQN.

    Warning:
        Your criterion must return a tensor of shape ``[batch_size,1]``
        when given two tensors of shape ``[batch_size,1]``, since we
        need to multiply the loss with importance sampling weight
        element-wise.

        If you are using loss modules given by pytorch. It is always
        safe to use them without any modification.
    """

    def __init__(self,
                 qnet: Union[NeuralNetworkModule, nn.Module],
                 qnet_target: Union[NeuralNetworkModule, nn.Module],
                 optimizer: Callable,
                 criterion: Callable,
                 *_,
                 lr_scheduler: Callable = None,
                 lr_scheduler_args: Tuple[Tuple] = None,
                 lr_scheduler_kwargs: Tuple[Dict] = None,
                 batch_size: int = 100,
                 update_rate: float = 0.005,
                 learning_rate: float = 0.001,
                 discount: float = 0.99,
                 gradient_max: float = np.inf,
                 replay_size: int = 500000,
                 replay_device: Union[str, t.device] = "cpu",
                 replay_buffer: Buffer = None,
                 visualize: bool = False,
                 visualize_dir: str = "",
                 **__):
        # DOC INHERITED
        super(DQNPer, self).__init__(
            qnet, qnet_target, optimizer, criterion,
            lr_scheduler=lr_scheduler,
            lr_scheduler_args=lr_scheduler_args,
            lr_scheduler_kwargs=lr_scheduler_kwargs,
            batch_size=batch_size,
            update_rate=update_rate,
            learning_rate=learning_rate,
            discount=discount,
            gradient_max=gradient_max,
            replay_size=replay_size,
            replay_device=replay_device,
            replay_buffer=(PrioritizedBuffer(replay_size, replay_device)
                           if replay_buffer is None
                           else replay_buffer),
            mode="double",
            visualize=visualize,
            visualize_dir=visualize_dir
        )
        # reduction must be None
        if not hasattr(self.criterion, "reduction"):
            raise RuntimeError("Criterion does not have the "
                               "'reduction' property")
        else:
            if hasattr(self.criterion, "reduction"):
                # A loss defined in ``torch.nn.modules.loss``
                if self.criterion.reduction != "none":
                    default_logger.warning(
                        "The reduction property of criterion is not 'none', "
                        "automatically corrected."
                    )
                    self.criterion.reduction = "none"

    def update(self,
               update_value=True,
               update_target=True,
               concatenate_samples=True,
               **__):
        # DOC INHERITED
        (batch_size,
         (state, action, reward, next_state, terminal, others),
         index, is_weight) = \
            self.replay_buffer.sample_batch(self.batch_size,
                                            concatenate_samples,
                                            sample_method="random_unique",
                                            sample_attrs=[
                                                "state", "action",
                                                "reward", "next_state",
                                                "terminal", "*"
                                            ])

        with t.no_grad():
            next_q_value = self._criticize(next_state)
            target_next_q_value = self._criticize(next_state, True)
            target_next_q_value = target_next_q_value.gather(
                dim=1, index=t.max(next_q_value, dim=1)[1].unsqueeze(1))

        q_value = self._criticize(state)

        # gather requires long tensor, int32 is not accepted
        action_value = q_value.gather(dim=1,
                                      index=self.action_get_function(action)
                                      .to(device=q_value.device,
                                          dtype=t.long))

        # Generate value reference :math: `y_i`.
        y_i = self.reward_function(
            reward, self.discount, target_next_q_value, terminal, others
        )
        value_loss = self.criterion(action_value, y_i.to(action_value.device))
        value_loss = (value_loss *
                      t.from_numpy(is_weight).view([batch_size, 1])
                      .to(value_loss.device))
        value_loss = value_loss.mean()

        abs_error = (t.sum(t.abs(action_value - y_i.to(action_value.device)),
                           dim=1)
                     .flatten().detach().cpu().numpy())
        self.replay_buffer.update_priority(abs_error, index)

        if self.visualize:
            self.visualize_model(value_loss, "qnet", self.visualize_dir)

        if update_value:
            self.qnet.zero_grad()
            value_loss.backward()
            nn.utils.clip_grad_norm_(
                self.qnet.parameters(), self.grad_max
            )
            self.qnet_optim.step()

        # Update target Q network
        if update_target:
            soft_update(self.qnet_target, self.qnet, self.update_rate)

        # use .item() to prevent memory leakage
        return value_loss.item()



from itertools import repeat
from typing import Tuple, Callable
from threading import Lock
from multiprocessing import get_context
import gym
import numpy as np

from machin.parallel.exception import ExceptionWithTraceback
from machin.parallel.queue import SimpleQueue
from machin.parallel.pickle import dumps, loads

from .base import *
from ..utils.openai_gym import disable_view_window

disable_view_window()


class GymTerminationError(Exception):
    def __init__(self):
        super(GymTerminationError, self).__init__(
            "One or several environments have terminated, "
            "reset before continuing."
        )


class ParallelWrapperDummy(ParallelWrapperBase):
    """
    Dummy parallel wrapper for gym environments, implemented using for-loop.

    For debug purpose only.
    """
    def __init__(self, env_creators: List[Callable[[int], gym.Env]]):
        """
        Args:
            env_creators: List of gym environment creators, used to create
                environments, accepts a index as your environment id.
        """
        super(ParallelWrapperDummy, self).__init__()
        self._envs = [ec(i) for ec, i in zip(env_creators,
                                             range(len(env_creators)))]
        self._terminal = np.zeros([len(self._envs)], dtype=np.bool)

    def reset(self, idx: Union[int, List[int]] = None) -> List[object]:
        """
        Returns:
            A list of gym states.
        """
        if idx is None:
            obsrv = [e.reset() for e in self._envs]
            self._terminal = np.zeros([self.size()], dtype=np.bool)
        else:
            obsrv = []
            if np.isscalar(idx):
                idx = [idx]
            for i in idx:
                obsrv.append(self._envs[i].reset())
                self._terminal[i] = False
        return obsrv

    def step(self,
             action: Union[np.ndarray, List[Any]],
             idx: Union[int, List[int]] = None) \
            -> Tuple[List[object], List[float], List[bool], List[dict]]:
        """
        Let specified environment(s) run one time step. Specified environments
        must be active and have not reached terminal states before.

        Args:
            action: Actions sent to each specified environment, the size of the
                first dimension must match the number of selected environments.
            idx: Indexes of selected environments, default is all.

        Returns:
            Observation, reward, terminal, and diagnostic info.
        """
        if idx is None:
            idx = list(range(self.size()))
        elif np.isscalar(idx):
            idx = [idx]

        if len(action) != len(idx):
            raise ValueError("Action number must match environment number!")
        if np.any(self._terminal[idx]):
            raise GymTerminationError()

        envs = [self._envs[i] for i in idx]
        result = [e.step(a) for e, a in zip(envs, action)]
        obsrv, reward, terminal, info = zip(*result)

        obsrv = list(obsrv)
        reward = list(reward)
        terminal = list(terminal)
        info = list(info)

        self._terminal[idx] |= terminal

        return obsrv, reward, terminal, info

    def seed(self, seed: Union[int, List[int]] = None) -> List[int]:
        """
        Set seeds for all environments.

        Args:
            seed: If seed is ``int``, the same seed will be used for all
                environments.
                If seed is ``List[int]``, it must have the same size as
                the number of all environments.
                If seed is ``None``, all environments will use the default
                seed.

        Returns:
            Actual used seed returned by all environments.
        """
        if np.isscalar(seed) or seed is None:
            seed = [seed] * self.size()
        result = []
        for e, s in zip(self._envs, seed):
            if hasattr(e, 'seed'):
                result.append(e.seed(s))
        return result

    def render(self, idx: Union[int, List[int]] = None,
               *_, **__) -> List[np.ndarray]:
        """
        Render all/specified environments.

        Args:
            idx: Indexes of selected environments, default is all.

        Returns:
            A list or rendered frames, of type ``np.ndarray`` and
            size (H, W, 3).
        """
        rendered = []
        if idx is None:
            for e in self._envs:
                if np.any(self._terminal):
                    raise GymTerminationError()
                rendered.append(e.render(mode="rgb_array"))
        else:
            if np.isscalar(idx):
                idx = [idx]
            for i in idx:
                if self._terminal[i]:
                    raise GymTerminationError()
                rendered.append(self._envs[i].render(mode="rgb_array"))
        return rendered

    def close(self) -> None:
        """
        Close all environments.
        """
        for e in self._envs:
            e.close()

    def active(self) -> List[int]:
        """
        Returns: Indexes of current active environments.
        """
        return np.arange(self.size())[~self._terminal]

    def size(self) -> int:
        """
        Returns: Number of environments.
        """
        return len(self._envs)

    @property
    def action_space(self) -> Any:
        # DOC INHERITED
        return self._envs[0].action_space

    @property
    def observation_space(self) -> Any:
        # DOC INHERITED
        return self._envs[0].observation_space


# noinspection PyBroadException
class ParallelWrapperSubProc(ParallelWrapperBase):
    """
    Parallel wrapper based on sub processes.
    """
    def __init__(self, env_creators: List[Callable[[int], gym.Env]]) -> None:
        """
        Args:
            env_creators: List of gym environment creators, used to create
                environments on sub process workers, accepts a index as your
                environment id.
        """
        super(ParallelWrapperSubProc, self).__init__()
        self.workers = []

        # Some environments will hang or collapse when using fork context.
        # E.g.: in "CarRacing-v0". pyglet used by gym will have render problems.

        # In case users wants to pass tensors to environments,
        # always copy all tensors to avoid errors
        ctx = get_context("spawn")
        self.cmd_queues = [SimpleQueue(ctx=ctx, copy_tensor=True)
                           for _ in range(len(env_creators))]
        self.result_queue = SimpleQueue(ctx=ctx, copy_tensor=True)
        for cmd_queue, ec, env_idx in zip(self.cmd_queues,
                                          env_creators,
                                          range(len(env_creators))):
            # enable recursive serialization to support
            # lambda & local function creators.
            self.workers.append(
                ctx.Process(target=self._worker,
                            args=(cmd_queue, self.result_queue,
                                  dumps(ec, recurse=True, copy_tensor=True),
                                  env_idx))
            )

        for worker in self.workers:
            worker.daemon = True
            worker.start()

        self.env_size = env_size = len(env_creators)
        self._cmd_lock = Lock()
        self._closed = False
        tmp_env = env_creators[0](0)
        self._action_space = tmp_env.action_space
        self._obsrv_space = tmp_env.observation_space
        tmp_env.close()
        self._terminal = np.zeros([env_size], dtype=np.bool)

    def reset(self, idx: Union[int, List[int]] = None) -> List[object]:
        """
        Returns:
            A list of gym states.
        """
        env_idxs = self._select_envs(idx)
        self._terminal[env_idxs] = False
        with self._cmd_lock:
            return self._call_gym_env_method(env_idxs, "reset")

    def step(self,
             action: Union[np.ndarray, List[Any]],
             idx: Union[int, List[int]] = None) \
            -> Tuple[List[object], List[float], List[bool], List[dict]]:
        """
        Let specified environment(s) run one time step. Specified environments
        must be active and have not reached terminal states before.

        Args:
            action: Actions sent to each specified environment, the size of the
                first dimension must match the number of selected environments.
            idx: Indexes of selected environments, default is all.

        Returns:
            Observation, reward, terminal, and diagnostic info.
        """
        env_idxs = self._select_envs(idx)
        if len(action) != len(env_idxs):
            raise ValueError("Action number must match environment number!")

        with self._cmd_lock:
            result = self._call_gym_env_method(env_idxs, "step",
                                               [(act,) for act in action])

        obsrv = [r[0] for r in result]
        reward = [r[1] for r in result]
        terminal = [r[2] for r in result]
        info = [r[3] for r in result]

        self._terminal[env_idxs] |= terminal

        return obsrv, reward, terminal, info

    def seed(self, seed: Union[int, List[int]] = None) -> List[int]:
        """
        Set seeds for all environments.

        Args:
            seed: If seed is ``int``, the same seed will be used for all
                environments.
                If seed is ``List[int]``, it must have the same size as
                the number of all environments.
                If seed is ``None``, all environments will use the default
                seed.

        Returns:
            Actual used seed returned by all environments.
        """
        if np.isscalar(seed) or seed is None:
            seed = [seed] * self.size()
        env_idxs = self._select_envs()
        with self._cmd_lock:
            return self._call_gym_env_method(env_idxs, "seed",
                                             [(sd,) for sd in seed])

    def render(self, idx: Union[int, List[int]] = None,
               *args, **kwargs) -> List[np.ndarray]:
        """
        Render all/specified environments.

        Args:
            idx: Indexes of selected environments, default is all.

        Returns:
            A list or rendered frames, of type ``np.ndarray`` and size
            (H, W, 3).
        """
        env_idxs = self._select_envs(idx)
        with self._cmd_lock:
            return self._call_gym_env_method(
                env_idxs, "render",
                kwargs=list(repeat({"mode": "rgb_array"}, len(env_idxs)))
            )

    def close(self) -> None:
        """
        Close all environments, including the wrapper.
        """
        with self._cmd_lock:
            if self._closed:
                return
            self._closed = True
            env_idxs = self._select_envs()
            self._call_gym_env_method(env_idxs, "close")
            for cmd_queue in self.cmd_queues:
                cmd_queue.quick_put(None)
            for worker in self.workers:
                worker.join()

    def active(self) -> List[int]:
        """
        Returns: Indexes of current active environments.
        """
        return np.arange(self.size())[~self._terminal]

    def size(self) -> int:
        """
        Returns: Number of environments.
        """
        return self.env_size

    @property
    def action_space(self) -> Any:
        # DOC INHERITED
        return self._action_space

    @property
    def observation_space(self) -> Any:
        # DOC INHERITED
        return self._obsrv_space

    def _select_envs(self, idx=None):
        if idx is None:
            idx = list(range(self.env_size))
        else:
            if np.isscalar(idx):
                idx = [idx]
        return idx

    def _call_gym_env_method(self, env_idxs, method, args=None, kwargs=None):
        if args is None:
            args = [() for _ in range(len(env_idxs))]
        if kwargs is None:
            kwargs = [{} for _ in range(len(env_idxs))]

        result = {}
        # Check whether any process has exited with error code:
        for worker, worker_id in zip(self.workers, range(len(self.workers))):
            if worker.exitcode is None:
                continue
            if worker.exitcode == 2:
                raise RuntimeError(
                    "Worker {} failed to create environment."
                    .format(worker_id)
                )
            elif worker.exitcode != 0:
                raise RuntimeError(
                    "Worker {} exited with code {}."
                    .format(worker_id, worker.exitcode)
                )

        for env_idx, i in zip(env_idxs, range(len(env_idxs))):
            self.cmd_queues[env_idx].quick_put((method, args[i], kwargs[i]))
        while len(result) < len(env_idxs):
            e_idx, success, res = self.result_queue.get()
            if success:
                result[e_idx] = res
            else:
                raise res
        return [result[e_idx] for e_idx in env_idxs]

    @staticmethod
    def _worker(cmd_queue: SimpleQueue,
                result_queue: SimpleQueue,
                env_creator, env_idx):
        env = None
        try:
            env = loads(env_creator)(env_idx)
        except Exception:
            # Something has gone wrong during environment creation,
            # exit with error.
            exit(2)
        try:
            while True:
                try:
                    command = cmd_queue.quick_get(timeout=1e-3)
                except TimeoutError:
                    continue

                try:
                    if command is not None:
                        method, args, kwargs = command
                    else:
                        # End of all tasks signal received
                        cmd_queue.close()
                        result_queue.close()
                        break
                    result = getattr(env, method)(*args, **kwargs)
                    result_queue.put((env_idx, True, result))
                except Exception as e:
                    # Something has gone wrong during execution, serialize
                    # the exception and send it back to master.
                    result_queue.put((env_idx, False,
                                      ExceptionWithTraceback(e)))
        except KeyboardInterrupt:
            cmd_queue.close()
            result_queue.close()



from . import ordered_server
from . import param_server
from .ordered_server import (
    OrderedServerBase,
    OrderedServerSimple,
    OrderedServerSimpleImpl
)
from .param_server import (
    PushPullGradServer,
    PushPullGradServerImpl,
    PushPullModelServer,
    PushPullModelServerImpl
)

__all__ = ["OrderedServerBase",
           "OrderedServerSimple",
           "OrderedServerSimpleImpl",
           "PushPullGradServer",
           "PushPullGradServerImpl",
           "PushPullModelServer",
           "PushPullModelServerImpl",
           "ordered_server", "param_server"]
